import math

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyQuantities, \
    EmptyExplainableObject
from efootprint.constants.units import u


def compute_nb_avg_hourly_occurrences(
        hourly_occurrences_starts: ExplainableHourlyQuantities, event_duration: ExplainableQuantity):
    if isinstance(hourly_occurrences_starts, EmptyExplainableObject):
        return EmptyExplainableObject()

    nb_avg_hourly_occurrences_in_parallel = None
    event_duration_in_nb_of_hours = event_duration.to(u.hour).magnitude
    nb_of_full_hours_in_event_duration = math.floor(event_duration_in_nb_of_hours)

    for hour_shift in range(0, nb_of_full_hours_in_event_duration):
        if nb_avg_hourly_occurrences_in_parallel is None:
            nb_avg_hourly_occurrences_in_parallel = hourly_occurrences_starts.value.shift(hour_shift, freq="h")
        else:
            nb_avg_hourly_occurrences_in_parallel = nb_avg_hourly_occurrences_in_parallel.add(
                hourly_occurrences_starts.value.shift(hour_shift, freq="h"), fill_value=0)

    nonfull_duration_rest = event_duration_in_nb_of_hours - nb_of_full_hours_in_event_duration
    if nonfull_duration_rest > 0:
        if nb_avg_hourly_occurrences_in_parallel is None:
            nb_avg_hourly_occurrences_in_parallel = hourly_occurrences_starts.value.shift(
                nb_of_full_hours_in_event_duration, freq="h") * nonfull_duration_rest
        else:
            nb_avg_hourly_occurrences_in_parallel = nb_avg_hourly_occurrences_in_parallel.add(
                hourly_occurrences_starts.value.shift(nb_of_full_hours_in_event_duration, freq="h")
                * nonfull_duration_rest,
                fill_value=0)
        
    return ExplainableHourlyQuantities(
            nb_avg_hourly_occurrences_in_parallel, left_parent=hourly_occurrences_starts, right_parent=event_duration,
            operator=f"hourly occurrences average")



import math
from typing import List, Type

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.service import Service
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.usage.compute_nb_occurrences_in_parallel import compute_nb_avg_hourly_occurrences


class JobTypes:
    AUTH = "auth"
    DATA_READ = "data_read"
    DATA_WRITE = "data_write"
    DATA_LIST = "data_list"
    DATA_SIMPLE_ANALYTIC = "data_simple_analytic"
    DATA_STREAM = "data_stream" #video, musique, data 
    TRANSACTION = "transaction"
    TRANSACTION_STRONG = "transaction_strong"
    NOTIFICATION = "notification"
    ANALYTIC_DATA_LOADING = "analytic_data_loading"
    ANALYTIC_READING_PREPARED = "analytic_reading_prepared"
    ANALYTIC_READING_ON_THE_FLY = "analytic_reading_on_the_fly"
    ML_RECOMMENDATION ="ml_reco" #kvm
    ML_LLM ="ml_llm"
    ML_DEEPLEARNING ="ml_dl"
    ML_REGRESSION = "ml_regression" #linear regression, polynomial regression, svm
    ML_CLASSIFIER = "ml_classifier" #bayes, random forest
    UNDEFINED = "undefined"


class Job(ModelingObject):
    def __init__(self, name: str, service: Service, data_upload: SourceValue, data_download: SourceValue,
                 request_duration: SourceValue, cpu_needed: SourceValue, ram_needed: SourceValue,
                 job_type: JobTypes = JobTypes.UNDEFINED, description: str = ""):
        super().__init__(name)
        self.hourly_occurrences_per_usage_pattern = ExplainableObjectDict()
        self.hourly_avg_occurrences_per_usage_pattern = ExplainableObjectDict()
        self.hourly_data_upload_per_usage_pattern = ExplainableObjectDict()
        self.hourly_data_download_per_usage_pattern = ExplainableObjectDict()
        self.hourly_occurrences_across_usage_patterns = None
        self.hourly_avg_occurrences_across_usage_patterns = None
        self.hourly_data_upload_across_usage_patterns = None
        self.job_type = job_type
        self.service = service
        if not data_upload.value.check("[]"):
            raise ValueError("Variable 'data_upload' does not have the appropriate '[]' dimensionality")
        self.data_upload = data_upload.set_label(f"Data upload of request {self.name}")
        if not data_download.value.check("[]"):
            raise ValueError("Variable 'data_upload' does not have the appropriate '[]' dimensionality")
        self.data_download = data_download.set_label(f"Data download of request {self.name}")
        if not request_duration.value.check("[time]"):
            raise ValueError("Variable 'request_duration' does not have the appropriate '[time]' dimensionality")
        self.request_duration = request_duration.set_label(f"Request duration to {self.service.name} in {self.name}")
        if not ram_needed.value.check("[]"):
            raise ValueError(
                "Variable 'ram_needed' does not have the appropriate '[]' dimensionality")
        self.ram_needed = ram_needed.set_label(
            f"RAM needed on server {self.service.server.name} to process {self.name}")
        if not cpu_needed.value.check("[cpu]"):
            raise ValueError(
                "Variable 'cpu_needed' does not have the appropriate '[cpu]' dimensionality")
        self.cpu_needed = cpu_needed.set_label(
            f"CPU needed on server {self.service.server.name} to process {self.name}")

        self.description = description
        
    @property
    def calculated_attributes(self) -> List[str]:
        return ["hourly_occurrences_per_usage_pattern", "hourly_avg_occurrences_per_usage_pattern",
                "hourly_data_upload_per_usage_pattern", "hourly_data_download_per_usage_pattern",
                "hourly_occurrences_across_usage_patterns", "hourly_avg_occurrences_across_usage_patterns",
                "hourly_data_upload_across_usage_patterns"]

    @property
    def duration_in_full_hours(self):
        return ExplainableQuantity(
                math.ceil(self.request_duration.to(u.hour).magnitude) * u.dimensionless,
                f"{self.name} duration in full hours")

    @property
    def user_journey_steps(self) -> List[Type["UserJourneyStep"]]:
        return self.modeling_obj_containers

    @property
    def user_journeys(self) -> List[Type["UserJourney"]]:
        return list(set(sum([uj_step.user_journeys for uj_step in self.user_journey_steps], start=[])))

    @property
    def usage_patterns(self) -> List[Type["UsagePattern"]]:
        return list(set(sum([uj_step.usage_patterns for uj_step in self.user_journey_steps], start=[])))

    @property
    def systems(self) -> List[Type["System"]]:
        return list(set(sum([up.systems for up in self.usage_patterns], start=[])))

    @property
    def networks(self) -> List[Type["Network"]]:
        return list(set(up.network for up in self.usage_patterns))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return [self.service] + self.networks

    def compute_hourly_occurrences_for_usage_pattern(self, usage_pattern: Type["UsagePattern"]):
        job_occurrences = EmptyExplainableObject()
        delay_between_uj_start_and_job_evt = EmptyExplainableObject()
        delay_in_hours_between_uj_start_and_job_evt = 0
        for uj_step in usage_pattern.user_journey.uj_steps:
            for uj_step_job in uj_step.jobs:
                if uj_step_job == self:
                    job_occurrences += usage_pattern.utc_hourly_user_journey_starts.return_shifted_hourly_quantities(
                        delay_in_hours_between_uj_start_and_job_evt)

            delay_between_uj_start_and_job_evt += uj_step.user_time_spent
            delay_in_hours_between_uj_start_and_job_evt = math.floor(
                delay_between_uj_start_and_job_evt.to(u.hour).magnitude)

        return job_occurrences.set_label(f"Hourly {self.name} occurrences in {usage_pattern.name}")

    def update_hourly_occurrences_per_usage_pattern(self):
        self.hourly_occurrences_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.hourly_occurrences_per_usage_pattern[up] = self.compute_hourly_occurrences_for_usage_pattern(up)

    def update_hourly_avg_occurrences_per_usage_pattern(self):
        self.hourly_avg_occurrences_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            hourly_avg_job_occurrences = compute_nb_avg_hourly_occurrences(
                self.hourly_occurrences_per_usage_pattern[up], self.request_duration)

            self.hourly_avg_occurrences_per_usage_pattern[up] = hourly_avg_job_occurrences.set_label(
                f"Average hourly {self.name} occurrences in {up.name}")

    def compute_hourly_data_exchange_for_usage_pattern(self, usage_pattern, data_exchange_type: str):
        data_exchange_type_no_underscore = data_exchange_type.replace("_", " ")

        hourly_data_exchange = EmptyExplainableObject()
        data_exchange_per_hour = (getattr(self, data_exchange_type) / self.duration_in_full_hours).set_label(
            f"{data_exchange_type_no_underscore} per hour for job {self.name} in {usage_pattern.name}")

        for hour_shift in range(0, self.duration_in_full_hours.magnitude):
            if not isinstance(self.hourly_occurrences_per_usage_pattern[usage_pattern], EmptyExplainableObject):
                hourly_data_exchange += (
                        self.hourly_occurrences_per_usage_pattern[usage_pattern].return_shifted_hourly_quantities(
                            hour_shift) * data_exchange_per_hour)

        return hourly_data_exchange.set_label(
                f"Hourly {data_exchange_type_no_underscore} for {self.name} in {usage_pattern.name}")

    def update_hourly_data_upload_per_usage_pattern(self):
        self.hourly_data_upload_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.hourly_data_upload_per_usage_pattern[up] = self.compute_hourly_data_exchange_for_usage_pattern(
                up, "data_upload")

    def update_hourly_data_download_per_usage_pattern(self):
        self.hourly_data_download_per_usage_pattern = ExplainableObjectDict()
        for up in self.usage_patterns:
            self.hourly_data_download_per_usage_pattern[up] = self.compute_hourly_data_exchange_for_usage_pattern(
                up, "data_download")
            
    def sum_calculated_attribute_across_usage_patterns(
            self, calculated_attribute_name: str, calculated_attribute_label: str):
        hourly_calc_attr_summed_across_ups = EmptyExplainableObject()
        for usage_pattern in self.usage_patterns:
            hourly_calc_attr_summed_across_ups += getattr(self, calculated_attribute_name)[usage_pattern]

        return hourly_calc_attr_summed_across_ups.set_label(
                f"Hourly {self.name} {calculated_attribute_label} across usage patterns")

    def update_hourly_occurrences_across_usage_patterns(self):
        self.hourly_occurrences_across_usage_patterns = self.sum_calculated_attribute_across_usage_patterns(
            "hourly_occurrences_per_usage_pattern", "occurrences")

    def update_hourly_avg_occurrences_across_usage_patterns(self):
        self.hourly_avg_occurrences_across_usage_patterns = self.sum_calculated_attribute_across_usage_patterns(
            "hourly_avg_occurrences_per_usage_pattern", "average occurrences")

    def update_hourly_data_upload_across_usage_patterns(self):
        self.hourly_data_upload_across_usage_patterns = self.sum_calculated_attribute_across_usage_patterns(
            "hourly_data_upload_per_usage_pattern", "data upload")


from typing import List

import pandas as pd

from efootprint.abstract_modeling_classes.source_objects import SourceValue, Source
from efootprint.core.usage.job import Job, JobTypes
from efootprint.core.service import Service
from efootprint.builders.usage.job_ecobenchmark.ecobenchmark_data_analysis import ECOBENCHMARK_DATA, \
    ECOBENCHMARK_RESULTS_LINK, default_request_duration
from efootprint.constants.units import u

ECOBENCHMARK_DF = pd.read_csv(ECOBENCHMARK_DATA)
ecobenchmark_source = Source(
    "e-footprint analysis of Boavizta’s Ecobenchmark data", ECOBENCHMARK_RESULTS_LINK)


def ecobenchmark_job(
        name: str, service: Service, data_upload: SourceValue, data_download: SourceValue,
        technology: str, implementation_details: str = "default", job_type=JobTypes.UNDEFINED, description: str = ""):
    filter_df = ECOBENCHMARK_DF[
        (ECOBENCHMARK_DF['service'] == technology) & (ECOBENCHMARK_DF['use_case'] == implementation_details)]

    if len(filter_df) == 0:
        raise ValueError(f"The input technology ({technology}) doesn’t belong to the list of supported technologies: "
                         f"{get_ecobenchmark_technologies()}")
    else:
        tech_row = filter_df.iloc[0]

    cpu_needed = SourceValue(tech_row['avg_cpu_core_per_request'] * u.core, ecobenchmark_source)
    ram_needed = SourceValue(tech_row['avg_ram_per_request_in_MB'] * u.MB, ecobenchmark_source)

    return Job(
        name, service, data_upload, data_download, request_duration=default_request_duration(),
        cpu_needed=cpu_needed, ram_needed=ram_needed, job_type=job_type, description=description)


def get_ecobenchmark_technologies() -> List[str]:
    return list(ECOBENCHMARK_DF["service"].unique())


if __name__ == "__main__":
    print(get_ecobenchmark_technologies())



from abc import abstractmethod
from typing import List

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyQuantities, ExplainableQuantity, \
    EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.sources import Sources, SOURCE_VALUE_DEFAULT_NAME
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


class Hardware(ModelingObject):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, fraction_of_usage_time: SourceValue):
        super().__init__(name)
        self.carbon_footprint_fabrication = carbon_footprint_fabrication.set_label(
            f"Carbon footprint fabrication of {self.name}")
        self.power = power.set_label(f"Power of {self.name}")
        self.lifespan = lifespan.set_label(f"Lifespan of {self.name}")
        if not fraction_of_usage_time.value.check("[]"):
            raise ValueError("Variable 'fraction_of_usage_per_day' shouldn’t have any dimensionality")
        self.fraction_of_usage_time = fraction_of_usage_time.set_label(f"{self.name} fraction of usage time")

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return self.modeling_obj_containers

    @property
    def systems(self) -> List:
        return list(set(sum([mod_obj.systems for mod_obj in self.modeling_obj_containers], start=[])))


class InfraHardware(Hardware):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue, lifespan: SourceValue,
                 average_carbon_intensity: SourceValue):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, SourceValue(1 * u.dimensionless, Sources.HYPOTHESIS))
        self.all_services_cpu_needs = None
        self.all_services_ram_needs = None
        self.raw_nb_of_instances = None
        self.nb_of_instances = None
        self.instances_energy = None
        self.energy_footprint = None
        self.instances_fabrication_footprint = None
        if not average_carbon_intensity.value.check("[time]**2 / [length]**2"):
            raise ValueError(
                "Variable 'average_carbon_intensity' does not have mass over energy "
                "('[time]**2 / [length]**2') dimensionality"
            )
        self.average_carbon_intensity = average_carbon_intensity
        if self.average_carbon_intensity.label == SOURCE_VALUE_DEFAULT_NAME:
            self.average_carbon_intensity.set_label(f"Average carbon intensity of {self.name} electricity")

    @property
    def calculated_attributes_defined_in_infra_hardware_class(self):
        return [
            "all_services_cpu_needs", "all_services_ram_needs", "raw_nb_of_instances",
            "nb_of_instances", "instances_fabrication_footprint", "instances_energy", "energy_footprint"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @abstractmethod
    def update_raw_nb_of_instances(self):
        pass

    @abstractmethod
    def update_nb_of_instances(self):
        pass

    @abstractmethod
    def update_instances_energy(self):
        pass

    @property
    def services(self):
        return self.modeling_obj_containers

    @property
    def systems(self) -> List:
        return list(set(sum([service.systems for service in self.services], start=[])))

    def update_all_services_ram_needs(self):
        all_services_ram_needs = EmptyExplainableObject()
        for service in self.services:
            all_services_ram_needs += service.hour_by_hour_ram_need

        self.all_services_ram_needs = all_services_ram_needs.to(u.GB).set_label(
                f"RAM needs of all services running on {self.name}")

    def update_all_services_cpu_needs(self):
        all_services_cpu_needs = EmptyExplainableObject()
        for service in self.services:
            all_services_cpu_needs += service.hour_by_hour_cpu_need

        self.all_services_cpu_needs = all_services_cpu_needs.to(u.core).set_label(
            f"CPU needs of all services running on {self.name}")

    def update_instances_fabrication_footprint(self):
        instances_fabrication_footprint = (
                self.carbon_footprint_fabrication * self.nb_of_instances * ExplainableQuantity(1 * u.hour, "one hour")
                / self.lifespan)

        self.instances_fabrication_footprint = instances_fabrication_footprint.to(u.kg).set_label(
                f"Hourly {self.name} instances fabrication footprint")

    def update_energy_footprint(self):
        energy_footprint = (self.instances_energy * self.average_carbon_intensity)

        self.energy_footprint = energy_footprint.to(u.kg).set_label(f"Hourly {self.name} energy footprint")



from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyQuantities
from efootprint.constants.sources import Sources


SOURCE_VALUE_DEFAULT_NAME = "unnamed source"


class SourceObject(ExplainableObject):
    def __init__(self, value: object, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)


class SourceValue(ExplainableQuantity):
    def __init__(self, value: Quantity, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)


class SourceHourlyValues(ExplainableHourlyQuantities):
    def __init__(self, value: Quantity, source: Source = Sources.HYPOTHESIS, label: str = SOURCE_VALUE_DEFAULT_NAME):
        super().__init__(value, label=label, source=source)



import logging
import os
from datetime import datetime

from efootprint.constants.files import create_folder, EFOOTPRINT_ROOT_PATH

logger = logging.getLogger('footprint-model')
logger.setLevel(logging.DEBUG)

ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


def write_logs_to_file(input_logger=logger, input_formatter=formatter, log_level=logging.INFO):
    log_path = create_folder(os.path.join(EFOOTPRINT_ROOT_PATH, "logs"))
    today_str = datetime.now().strftime("%Y-%m-%d")
    log_file = os.path.join(log_path, today_str + ".log")
    fh = logging.FileHandler(log_file)
    fh.setFormatter(input_formatter)
    fh.setLevel(log_level)
    input_logger.addHandler(fh)


if os.environ.get("WRITE_EFOOTPRINT_LOGS", None) is not None:
    write_logs_to_file(log_level=logging.DEBUG)



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.network import Network


def default_wifi_network(name="Default wifi network", **kwargs):
    output_args = {
        "bandwidth_energy_intensity": SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY)
    }

    output_args.update(kwargs)

    return Network(name, **output_args)


def default_mobile_network(name="Default mobile network", **kwargs):
    output_args = {
        "bandwidth_energy_intensity": SourceValue(0.12 * u("kWh/GB"), Sources.TRAFICOM_STUDY)
    }

    output_args.update(kwargs)

    return Network(name, **output_args)



from typing import Type, Optional
from abc import ABC, abstractmethod
from dataclasses import dataclass
import os
import json

from IPython.display import HTML

from efootprint.logger import logger
from efootprint.utils.calculus_graph import build_calculus_graph
from efootprint.utils.graph_tools import add_unique_id_to_mynetwork


class ObjectLinkedToModelingObj(ABC):
    def __init__(self):
        self.modeling_obj_container = None
        self.attr_name_in_mod_obj_container = None

    @abstractmethod
    def set_modeling_obj_container(self, new_parent_modeling_object: Type["ModelingObject"], attr_name: str):
        pass


@dataclass
class Source:
    name: str
    link: Optional[str]


class ExplainableObject(ObjectLinkedToModelingObj):
    def __init__(
            self, value: object, label: str = None, left_parent: Type["ExplainableObject"] = None,
            right_parent: Type["ExplainableObject"] = None, operator: str = None, source: Source = None):
        super().__init__()
        self.value = value
        if not label and (left_parent is None and right_parent is None):
            raise ValueError(f"ExplainableObject without parent should have a label")
        if source is not None and (left_parent is not None or right_parent is not None):
            raise ValueError(f"An ExplainableObject with a source shouldn’t have any parent")
        self.source = source
        self.label = None
        self.set_label(label)
        self.left_parent = left_parent
        self.right_parent = right_parent
        self.operator = operator
        self.direct_ancestors_with_id = []
        self.direct_children_with_id = []

        for parent in (self.left_parent, self.right_parent):
            if parent is not None:
                self.direct_ancestors_with_id += [
                    ancestor_with_id for ancestor_with_id in parent.return_direct_ancestors_with_id_to_child()
                    if ancestor_with_id.id not in self.direct_ancestor_ids]

    def __deepcopy__(self, memo):
        cls = self.__class__
        new_instance = cls.__new__(cls)
        new_instance.__init__(value=self.value, label=self.label, source=getattr(self, "source", None))

        return new_instance

    @property
    def id(self):
        if self.modeling_obj_container is None:
            raise ValueError(
                f"{self.label} doesn’t have a modeling_obj_container, hence it makes no sense "
                f"to look for its ancestors")
        return f"{self.attr_name_in_mod_obj_container}-in-{self.modeling_obj_container.id}"

    @property
    def has_parent(self):
        return self.left_parent is not None or self.right_parent is not None

    @property
    def direct_ancestor_ids(self):
        return [attr.id for attr in self.direct_ancestors_with_id]

    @property
    def direct_child_ids(self):
        return [attr.id for attr in self.direct_children_with_id]

    def set_modeling_obj_container(self, new_modeling_obj_container: Type["ModelingObject"], attr_name: str):
        if not self.label:
            raise ValueError(f"ExplainableObjects that are attributes of a ModelingObject should always have a label.")
        if self.modeling_obj_container is not None and new_modeling_obj_container.id != self.modeling_obj_container.id:
            logger.warning(
                f"Linking {self.label} to {new_modeling_obj_container.name}, erasing its existing link to "
                f"{self.modeling_obj_container.name}.")
            if self.left_parent is not None or self.right_parent is not None:
                raise ValueError(
                    f"An ExplainableObject with parent can’t be attributed to more than one ModelingObject. Here "
                    f"{self.label} is trying to be linked to {new_modeling_obj_container.name} but is already linked to"
                    f" {self.modeling_obj_container.name}."
                    f" A common reason why this error could happen is that a mutable object (SourceValue for"
                    f" example) has been set as default value in one of the classes.")
        self.modeling_obj_container = new_modeling_obj_container
        self.attr_name_in_mod_obj_container = attr_name
        for direct_ancestor_with_id in self.direct_ancestors_with_id:
            direct_ancestor_with_id.update_direct_children_with_id(direct_child=self)

    def return_direct_ancestors_with_id_to_child(self):
        if self.modeling_obj_container is not None:
            return [self]
        else:
            return self.direct_ancestors_with_id

    def update_direct_children_with_id(self, direct_child):
        if direct_child.id not in self.direct_child_ids:
            self.direct_children_with_id.append(direct_child)

    def set_label(self, new_label):
        if self.source is not None and f"from {self.source.name}" not in new_label:
            self.label = f"{new_label} from {self.source.name}"
        else:
            self.label = new_label

        return self

    def get_all_descendants_with_id(self):
        all_descendants = []

        def retrieve_descendants(expl_obj: ExplainableObject, descendants_list):
            for child in expl_obj.direct_children_with_id:
                if child.id not in [elt.id for elt in descendants_list]:
                    descendants_list.append(child)
                retrieve_descendants(child, descendants_list)

        retrieve_descendants(self, all_descendants)

        return all_descendants

    def explain(self, pretty_print=True):
        element_value_to_print = str(self)

        if self.left_parent is None and self.right_parent is None:
            return f"{self.label} = {element_value_to_print}"
        explain_tuples = self.compute_explain_nested_tuples()

        if pretty_print:
            return self.pretty_print_calculation(
                f"{self.label} = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=False)}"
                f" = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=True)}"
                f" = {element_value_to_print}")
        else:
            return f"{self.label} = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=False)}" \
                f" = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=True)}" \
                f" = {element_value_to_print}"

    def compute_explain_nested_tuples(self, return_label_if_self_has_one=False):
        if return_label_if_self_has_one and self.label:
            return self

        left_explanation = None
        right_explanation = None

        if self.left_parent:
            left_explanation = self.left_parent.compute_explain_nested_tuples(return_label_if_self_has_one=True)
        if self.right_parent:
            right_explanation = self.right_parent.compute_explain_nested_tuples(return_label_if_self_has_one=True)

        if left_explanation is None and right_explanation is None:
            raise ValueError("Object to explain should have at least one child")

        return left_explanation, self.operator, right_explanation

    def print_tuple_element(self, tuple_element: object, print_values_instead_of_labels: bool):
        if issubclass(type(tuple_element), ExplainableObject):
            if print_values_instead_of_labels:
                return str(tuple_element)
            else:
                return f"{tuple_element.label}"
        elif type(tuple_element) == str:
            return tuple_element
        elif type(tuple_element) == tuple:
            if tuple_element[1] is None:
                return f"{self.print_tuple_element(tuple_element[0], print_values_instead_of_labels)}"
            if tuple_element[2] is None:
                return f"{tuple_element[1]}" \
                       f" of ({self.print_tuple_element(tuple_element[0], print_values_instead_of_labels)})"

            left_parenthesis = False
            right_parenthesis = False

            if tuple_element[1] == "/":
                if type(tuple_element[2]) == tuple:
                    right_parenthesis = True
                if type(tuple_element[0]) == tuple and tuple_element[0][1] != "*":
                    left_parenthesis = True
            elif tuple_element[1] == "*":
                if type(tuple_element[0]) == tuple and tuple_element[0][1] != "*":
                    left_parenthesis = True
                if type(tuple_element[2]) == tuple and tuple_element[2][1] != "*":
                    right_parenthesis = True
            elif tuple_element[1] == "-":
                if type(tuple_element[2]) == tuple and tuple_element[2][1] in ["+", "-"]:
                    right_parenthesis = True
            elif tuple_element[1] == "+":
                pass

            lp_open = ""
            lp_close = ""
            rp_open = ""
            rp_close = ""

            if left_parenthesis:
                lp_open = "("
                lp_close = ")"
            if right_parenthesis:
                rp_open = "("
                rp_close = ")"

            return f"{lp_open}{self.print_tuple_element(tuple_element[0], print_values_instead_of_labels)}{lp_close}" \
                   f" {tuple_element[1]}" \
                   f" {rp_open}{self.print_tuple_element(tuple_element[2], print_values_instead_of_labels)}{rp_close}"

    @staticmethod
    def pretty_print_calculation(calc_str):
        return calc_str.replace(" = ", "\n=\n")

    def calculus_graph_to_file(
            self, filename=None, colors_dict=None, x_multiplier=150, y_multiplier=150, width="1800px", height="900px",
            notebook=False, max_depth=100):
        if colors_dict is None:
            colors_dict = {"user data": "gold", "default": "darkred"}
        calculus_graph = build_calculus_graph(
            self, colors_dict, x_multiplier, y_multiplier, width, height, notebook, max_depth=max_depth)

        if filename is None:
            filename = os.path.join(".", f"{self.label} calculus graph.html")

        calculus_graph.show(filename, notebook=notebook)

        add_unique_id_to_mynetwork(filename)

        if notebook:
            return HTML(filename)

    def to_json(self, with_calculated_attributes_data=False):
        output_dict = {"label": self.label}

        if type(self.value) == list:  # Case of time_intervals in UsagePattern class
            output_dict["value"] = self.value
        elif getattr(self.value, "zone", None) is not None:  # Case of timezone in Country class
            output_dict["zone"] = self.value.zone

        if self.source is not None:
            output_dict["source"] = {"name": self.source.name, "link": self.source.link}

        if with_calculated_attributes_data:
            output_dict["id"] = self.id
            output_dict["direct_ancestors_with_id"] = [elt.id for elt in self.direct_ancestors_with_id]
            output_dict["direct_children_with_id"] = [elt.id for elt in self.direct_children_with_id]

        return output_dict

    def __repr__(self):
        return json.dumps(self.to_json())

    def __str__(self):
        return str(self.value)



from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.servers.server_base_class import Server


class Serverless(Server):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, ram: SourceValue, cpu_cores: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 server_utilization_rate: SourceValue):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, idle_power, ram, cpu_cores, power_usage_effectiveness,
            average_carbon_intensity, server_utilization_rate)

    def update_nb_of_instances(self):
        hour_by_hour_nb_of_instances = self.raw_nb_of_instances.copy()

        self.nb_of_instances = hour_by_hour_nb_of_instances.set_label(f"Hour number of {self.name} instances")


from efootprint.abstract_modeling_classes.explainable_object_base_class import (
    ObjectLinkedToModelingObj, ExplainableObject)

from typing import Type
import json

from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject


class ExplainableObjectDict(ObjectLinkedToModelingObj, dict):
    # TODO: optimization opportunity: as such in the case of several values of an ExplainableObjectDict
    # depending on the same SourceValue if the SourceValue is updated the ExplainableObjectDict will be recomputed as
    # many times as it has values depending on the SourceValue.
    def __init__(self):
        super().__init__()
        self.modeling_obj_container = None
        self.attr_name_in_mod_obj_container = None

    def set_modeling_obj_container(self, new_parent_modeling_object: Type["ModelingObject"], attr_name: str):
        if self.modeling_obj_container is not None and new_parent_modeling_object.id != self.modeling_obj_container.id:
            raise ValueError(
                f"An ExplainableObjectDict can’t be attributed to more than one ModelingObject. Here "
                f"{self.label} is trying to be linked to {new_parent_modeling_object.name} but is already linked to "
                f"{self.modeling_obj_container.name}.")
        self.modeling_obj_container = new_parent_modeling_object
        self.attr_name_in_mod_obj_container = attr_name

    def __setitem__(self, key, value: ExplainableObject):
        if not issubclass(type(value), ExplainableObject) and not isinstance(value, EmptyExplainableObject):
            raise ValueError(
                f"ExplainableObjectDicts only accept ExplainableObjects or EmptyExplainableObject as values, "
                f"received {type(value)}")
        super().__setitem__(key, value)
        if value != 0:
            value.set_modeling_obj_container(
                new_modeling_obj_container=self.modeling_obj_container, attr_name=self.attr_name_in_mod_obj_container)

    def to_json(self, with_calculated_attributes_data=False):
        output_dict = {}

        for key, value in self.items():
            output_dict[key.id] = value.to_json(with_calculated_attributes_data)

        return output_dict

    def __repr__(self):
        return json.dumps(self.to_json())

    def __str__(self):
        return_str = "{\n"

        for key, value in self.items():
            return_str += f"{key.id}: {value}, \n"

        return_str = return_str + "}"

        return return_str



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.hardware_base_classes import Hardware


def default_smartphone(name="Default smartphone", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(30 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(1 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(3 * u.year, Sources.HYPOTHESIS),
        "fraction_of_usage_time": SourceValue(3.6 * u.hour / u.day, Sources.STATE_OF_MOBILE_2022)
    }

    output_args.update(kwargs)

    return Hardware(name, **output_args)


def default_laptop(name="Default laptop", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(156 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(50 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(6 * u.year, Sources.HYPOTHESIS),
        "fraction_of_usage_time": SourceValue(7 * u.hour / u.day, Sources.STATE_OF_MOBILE_2022)
    }

    output_args.update(kwargs)

    return Hardware(name, **output_args)


def default_box(name="Default box", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(78 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(10 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(6 * u.year, Sources.HYPOTHESIS),
        "fraction_of_usage_time": SourceValue(24 * u.hour / u.day, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return Hardware(name, **output_args)


def default_screen(name="Default screen", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(222 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(30 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(6 * u.year, Sources.HYPOTHESIS),
        "fraction_of_usage_time": SourceValue(7 * u.hour / u.day, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return Hardware(name, **output_args)



from abc import abstractmethod

import numpy as np
import pandas as pd
import pint_pandas

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyQuantities, ExplainableQuantity, \
    EmptyExplainableObject
from efootprint.core.hardware.hardware_base_classes import InfraHardware
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


class Server(InfraHardware):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, ram: SourceValue, cpu_cores: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 server_utilization_rate: SourceValue):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, average_carbon_intensity)
        self.available_cpu_per_instance = None
        self.available_ram_per_instance = None
        self.server_utilization_rate = None
        self.raw_nb_of_instances = None
        self.nb_of_instances = None
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.ram = ram.set_label(f"RAM of {self.name}")
        self.cpu_cores = cpu_cores.set_label(f"Nb cpus cores of {self.name}")
        self.power_usage_effectiveness = power_usage_effectiveness.set_label(f"PUE of {self.name}")
        self.server_utilization_rate = server_utilization_rate.set_label(f"{self.name} utilization rate")

    @property
    def calculated_attributes(self):
        return ["available_ram_per_instance", "available_cpu_per_instance"
                                      ] + self.calculated_attributes_defined_in_infra_hardware_class

    def update_available_ram_per_instance(self):
        services_base_ram_consumptions = [service.base_ram_consumption for service in self.services]
        available_ram_per_instance = self.ram * self.server_utilization_rate
        for service_base_resource_consumption in services_base_ram_consumptions:
            available_ram_per_instance -= service_base_resource_consumption
            if available_ram_per_instance.value < 0 * u.B:
                services_resource_need = sum(services_base_ram_consumptions)
                raise ValueError(
                    f"server has available capacity of "
                    f"{(self.ram * self.server_utilization_rate).value} "
                    f" but is asked {services_resource_need.value}")

        self.available_ram_per_instance = available_ram_per_instance.set_label(
            f"Available RAM per {self.name} instance")

    def update_available_cpu_per_instance(self):
        services_base_cpu_consumptions = [service.base_cpu_consumption for service in self.services]
        available_cpu_per_instance = self.cpu_cores * self.server_utilization_rate
        for service_base_resource_consumption in services_base_cpu_consumptions:
            available_cpu_per_instance -= service_base_resource_consumption
            if available_cpu_per_instance.value < 0:
                services_resource_need = sum(services_base_cpu_consumptions)
                raise ValueError(
                    f"server has available capacity of "
                    f"{(self.cpu_cores * self.server_utilization_rate).value} "
                    f" but is asked {services_resource_need.value}")

        self.available_cpu_per_instance = available_cpu_per_instance.set_label(
            f"Available CPU per {self.name} instance")

    def update_raw_nb_of_instances(self):
        if isinstance(self.all_services_ram_needs, EmptyExplainableObject) \
                and isinstance(self.all_services_cpu_needs, EmptyExplainableObject):
            self.raw_nb_of_instances = EmptyExplainableObject()
        else:
            nb_of_servers_based_on_ram_alone = (
                    self.all_services_ram_needs / self.available_ram_per_instance).to(u.dimensionless).set_label(
                f"Raw nb of {self.name} instances based on RAM alone")
            nb_of_servers_based_on_cpu_alone = (
                    self.all_services_cpu_needs / self.available_cpu_per_instance).to(u.dimensionless).set_label(
                f"Raw nb of {self.name} instances based on CPU alone")

            nb_of_servers_raw_np = np.maximum(
                nb_of_servers_based_on_ram_alone.value["value"].values.data,
                nb_of_servers_based_on_cpu_alone.value["value"].values.data
            )
            nb_of_servers_raw_df = pd.DataFrame(
                {"value": pint_pandas.PintArray(nb_of_servers_raw_np, dtype=u.dimensionless)},
                index=nb_of_servers_based_on_ram_alone.value.index)

            nb_of_servers_raw = ExplainableHourlyQuantities(
                nb_of_servers_raw_df,
                f"Raw nb of instances",
                left_parent=nb_of_servers_based_on_ram_alone,
                right_parent=nb_of_servers_based_on_cpu_alone,
                operator="max compared with"
            )
            hour_by_hour_raw_nb_of_instances = nb_of_servers_raw.set_label(
                f"Hourly raw number of {self.name} instances")

            self.raw_nb_of_instances = hour_by_hour_raw_nb_of_instances

    def update_instances_energy(self):
        energy_spent_by_one_idle_instance_over_one_hour = (
                self.idle_power * self.power_usage_effectiveness * ExplainableQuantity(1 * u.hour, "one hour"))
        extra_energy_spent_by_one_fully_active_instance_over_one_hour = (
                (self.power - self.idle_power) * self.power_usage_effectiveness
                * ExplainableQuantity(1 * u.hour, "one hour"))

        server_power = (
                energy_spent_by_one_idle_instance_over_one_hour * self.nb_of_instances
                + extra_energy_spent_by_one_fully_active_instance_over_one_hour * self.raw_nb_of_instances)

        self.instances_energy = server_power.to(u.kWh).set_label(
            f"Hourly energy consumed by {self.name} instances")


    @abstractmethod
    def update_nb_of_instances(self):
        pass


from typing import List

import pytz

from efootprint.constants.units import u
from efootprint.constants.sources import Source,Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject


class Country(ModelingObject):
    def __init__(
            self, name: str, short_name: str, average_carbon_intensity: SourceValue, year: int, timezone: SourceObject):
        super().__init__(name)
        self.short_name = short_name
        # "[time]**2 / [length]**2" corresponds to mass over energy I.U.
        if not average_carbon_intensity.value.check("[time]**2 / [length]**2"):
            raise ValueError(
                "Variable 'average_carbon_intensity' does not have mass over energy "
                "('[time]**2 / [length]**2') dimensionality"
            )
        self.average_carbon_intensity = average_carbon_intensity.set_label(f"Average carbon intensity of {self.name}")
        self.year = year
        self.timezone = timezone.set_label(f"{self.name} timezone")

    @property
    def usage_patterns(self):
        return self.modeling_obj_containers

    @property
    def systems(self) -> List:
        return list(set(sum([usage_pattern.systems for usage_pattern in self.usage_patterns], start=[])))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self):
        return self.usage_patterns


def tz(timezone: str):
    return SourceObject(pytz.timezone(timezone), Sources.USER_DATA)


def country_generator(country_name, country_short_name, country_avg_carbon_int, year, timezone):
    def return_country():
        return Country(country_name, country_short_name, country_avg_carbon_int, year, timezone)

    return return_country


class Countries:
    # TODO: Add other countries and automate data retrieval
    source = Source("Our world in data", "https://ourworldindata.org/energy#country-profiles")
    FRANCE = country_generator("France", "FRA", SourceValue(85 * u.g / u.kWh, source), 2022, tz('Europe/Paris'))
    # EUROPE = country_generator("Europe", "EUR", SourceValue(278 * u.g / u.kWh, source), 2022, None)
    GERMANY = country_generator("Germany", "DEU", SourceValue(386 * u.g / u.kWh, source), 2022, tz('Europe/Berlin'))
    FINLAND = country_generator("Finland", "FIN", SourceValue(132 * u.g / u.kWh, source), 2022, tz('Europe/Helsinki'))
    AUSTRIA = country_generator("Austria", "AUT", SourceValue(158 * u.g / u.kWh, source), 2022, tz('Europe/Vienna'))
    POLAND = country_generator("Poland", "POL", SourceValue(635 * u.g / u.kWh, source), 2022, tz('Europe/Warsaw'))
    NORWAY = country_generator("Norway", "NOR", SourceValue(26 * u.g / u.kWh, source), 2021, tz('Europe/Oslo'))
    HUNGARY = country_generator("Hungary", "HUN", SourceValue(223 * u.g / u.kWh, source), 2022, tz('Europe/Budapest'))
    UNITED_KINGDOM = country_generator("United Kingdom", "GBR", SourceValue(268 * u.g / u.kWh, source), 2021, tz('Europe/London'))
    BELGIUM = country_generator("Belgium", "BEL", SourceValue(165 * u.g / u.kWh, source), 2022, tz('Europe/Brussels'))
    ITALY = country_generator("Italy", "IT", SourceValue(371 * u.g / u.kWh, source), 2022, tz('Europe/Rome'))
    ROMANIA = country_generator("Romania", "RO", SourceValue(264 * u.g / u.kWh, source), 2022, tz('Europe/Bucharest'))
    MALAYSIA = country_generator("Malaysia", "MY", SourceValue(549 * u.g / u.kWh, source), 2021, tz('Asia/Kuala_Lumpur'))
    MOROCCO = country_generator("Morocco", "MA", SourceValue(610 * u.g / u.kWh, source), 2021, tz('Africa/Casablanca'))
    TUNISIA = country_generator("Tunisia", "TN", SourceValue(468 * u.g / u.kWh, source), 2021, tz('Africa/Tunis'))
    ALGERIA = country_generator("Algeria", "DZ", SourceValue(488 * u.g / u.kWh, source), 2021, tz('Africa/Algiers'))
    SENEGAL = country_generator("Senegal", "SN", SourceValue(503 * u.g / u.kWh, source), 2021, tz('Africa/Dakar'))
    # UNITED_STATES = country_generator("United States", "US", SourceValue(379 * u.g / u.kWh, source), 2021, None)
    # BRAZIL = country_generator("Brazil", "BR", SourceValue(159 * u.g / u.kWh, source), 2021, None)



from datetime import datetime

import numpy as np
import pandas as pd
import pint_pandas
import pytz
import json
from copy import copy

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.modeling_object import PREVIOUS_LIST_VALUE_SET_SUFFIX
from efootprint.abstract_modeling_classes.source_objects import SourceObject
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source
from efootprint.builders.time_builders import create_hourly_usage_df_from_list
from efootprint.constants.units import u
from efootprint.logger import logger
from efootprint.core.service import Service
from efootprint.core.system import System
from efootprint.core.hardware.storage import Storage
from efootprint.core.hardware.servers.autoscaling import Autoscaling
from efootprint.core.hardware.servers.serverless import Serverless
from efootprint.core.hardware.servers.on_premise import OnPremise
from efootprint.core.hardware.hardware_base_classes import Hardware
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.usage.user_journey import UserJourney
from efootprint.core.usage.job import Job
from efootprint.core.usage.user_journey_step import UserJourneyStep
from efootprint.core.hardware.network import Network
from efootprint.constants.countries import Country


def json_to_explainable_object(input_dict):
    output = None
    source = None
    if "source" in input_dict.keys():
        source = Source(input_dict["source"]["name"], input_dict["source"]["link"])
    if "value" in input_dict.keys() and "unit" in input_dict.keys():
        value = input_dict["value"] * u(input_dict["unit"])
        output = ExplainableQuantity(
            value, label=input_dict["label"], source=source)
    elif "values" in input_dict.keys() and "unit" in input_dict.keys():
        output = ExplainableHourlyQuantities(
            create_hourly_usage_df_from_list(
                input_dict["values"],
                pint_unit=u(input_dict["unit"]),
                start_date=datetime.strptime(input_dict["start_date"], "%Y-%m-%d %H:%M:%S"),
            ),
            label=input_dict["label"], source=source)
    elif "zone" in input_dict.keys():
        output = SourceObject(
            pytz.timezone(input_dict["zone"]), source, input_dict["label"])

    return output


def json_to_system(system_dict):
    class_obj_dict = {}
    flat_obj_dict = {}

    for class_key in system_dict.keys():
        if class_key not in class_obj_dict.keys():
            class_obj_dict[class_key] = {}
        current_class = globals()[class_key]
        current_class_dict = {}
        for class_instance_key in system_dict[class_key].keys():
            new_obj = current_class.__new__(current_class)
            new_obj.__dict__["modeling_obj_containers"] = []
            for attr_key, attr_value in system_dict[class_key][class_instance_key].items():
                if type(attr_value) == dict:
                    new_obj.__dict__[attr_key] = json_to_explainable_object(attr_value)
                    new_obj.__dict__[attr_key].set_modeling_obj_container(new_obj, attr_key)
                else:
                    new_obj.__dict__[attr_key] = attr_value

            current_class_dict[class_instance_key] = new_obj
            flat_obj_dict[class_instance_key] = new_obj

        class_obj_dict[class_key] = current_class_dict

    for class_key in class_obj_dict.keys():
        for mod_obj_key, mod_obj in class_obj_dict[class_key].items():
            for attr_key, attr_value in list(mod_obj.__dict__.items()):
                if type(attr_value) == str and attr_key != "id" and attr_value in flat_obj_dict.keys():
                    mod_obj.__dict__[attr_key] = flat_obj_dict[attr_value]
                    flat_obj_dict[attr_value].add_obj_to_modeling_obj_containers(mod_obj)
                elif type(attr_value) == list and attr_key != "modeling_obj_containers":
                    output_val = []
                    for elt in attr_value:
                        if type(elt) == str and elt in flat_obj_dict.keys():
                            output_val.append(flat_obj_dict[elt])
                            flat_obj_dict[elt].add_obj_to_modeling_obj_containers(mod_obj)
                    mod_obj.__dict__[attr_key] = output_val
                    mod_obj.__dict__[f"{attr_key}{PREVIOUS_LIST_VALUE_SET_SUFFIX}"] = copy(output_val)
            mod_obj.__dict__["dont_handle_input_updates"] = False
            mod_obj.__dict__["init_has_passed"] = True

    for obj_type in class_obj_dict.keys():
        if obj_type != "System":
            for mod_obj in class_obj_dict[obj_type].values():
                if len(mod_obj.systems) == 0:
                    logger.warning(
                        f"{mod_obj.class_as_simple_str} {mod_obj.name} is not linked to any existing system so needs "
                        f"to compute its own calculated attributes")
                    mod_obj.compute_calculated_attributes()

    for system in class_obj_dict["System"].values():
        system_id = system.id
        system.__init__(system.name, usage_patterns=system.usage_patterns)
        system.id = system_id
        system.after_init()

    return class_obj_dict, flat_obj_dict


def get_obj_by_key_similarity(obj_container_dict, input_key):
    for key in obj_container_dict.keys():
        if input_key in key:
            return obj_container_dict[key]



import json
import numbers
from datetime import datetime
from typing import Type

import pandas as pd
import pint_pandas
import pytz
from pint import Quantity, Unit
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates

from efootprint.abstract_modeling_classes.explainable_object_base_class import (
    ExplainableObject, Source, ObjectLinkedToModelingObj)
from efootprint.constants.units import u


class EmptyExplainableObject(ObjectLinkedToModelingObj):
    def set_modeling_obj_container(self, new_modeling_obj_container: Type["ModelingObject"], attr_name: str):
        self.modeling_obj_container = new_modeling_obj_container
        self.attr_name_in_mod_obj_container = attr_name

    def to(self, unit):
        return self

    def set_label(self, label):
        return self

    def ceil(self):
        return EmptyExplainableObject()

    def max(self):
        return EmptyExplainableObject()

    def abs(self):
        return EmptyExplainableObject()

    def sum(self):
        return EmptyExplainableObject()

    def copy(self):
        return EmptyExplainableObject()

    @property
    def magnitude(self):
        return self

    @property
    def value(self):
        return self

    def __eq__(self, other):
        if isinstance(other, EmptyExplainableObject):
            return True
        elif other == 0:
            return True

        return False

    def __add__(self, other):
        if issubclass(type(other), ExplainableObject):
            return other.__add__(self)
        elif isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject()
        elif other == 0:
            return self
        else:
            raise ValueError

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject()
        else:
            raise ValueError


class ExplainableQuantity(ExplainableObject):
    def __init__(
            self, value: Quantity, label: str = None, left_parent: Type["ExplainableQuantity"] = None,
            right_parent: Type["ExplainableQuantity"] = None, operator: str = None, source: Source = None):
        if not isinstance(value, Quantity):
            raise ValueError(
                f"Variable 'value' of type {type(value)} does not correspond to the appropriate 'Quantity' type, "
                "it is indeed mandatory to define a unit"
            )
        super().__init__(value, label, left_parent, right_parent, operator, source)

    def to(self, unit_to_convert_to):
        self.value = self.value.to(unit_to_convert_to)

        return self

    @property
    def magnitude(self):
        return self.value.magnitude

    def compare_with_and_return_max(self, other):
        if issubclass(type(other), ExplainableQuantity):
            if self.value >= other.value:
                return ExplainableQuantity(self.value, left_parent=self, right_parent=other, operator="max")
            else:
                return ExplainableQuantity(other.value, left_parent=self, right_parent=other, operator="max")
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def ceil(self):
        self.value = np.ceil(self.value)
        return self

    def __gt__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return self.value > other.value
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __lt__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return self.value < other.value
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __eq__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return self.value == other.value
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __add__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            # summing with sum() adds an implicit 0 as starting value
            return ExplainableQuantity(self.value, left_parent=self)
        elif isinstance(other, EmptyExplainableObject):
            return ExplainableQuantity(self.value, left_parent=self)
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value + other.value, "", self, other, "+")
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __sub__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return ExplainableQuantity(self.value, left_parent=self)
        elif isinstance(other, EmptyExplainableObject):
            return ExplainableQuantity(self.value, left_parent=self)
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value - other.value, "", self, other, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __mul__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return 0
        elif isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject()
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value * other.value, "", self, other, "*")
        elif issubclass(type(other), ExplainableHourlyQuantities):
            return other.__mul__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __truediv__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value / other.value, "", self, other, "/")
        elif issubclass(type(other), ExplainableHourlyQuantities):
            return other.__rtruediv__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(other.value - self.value, "", other, self, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rtruediv__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return 0
        elif isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject()
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(other.value / self.value, "", other, self, "/")
        elif issubclass(type(other), ExplainableHourlyQuantities):
            return other.__truediv__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __round__(self, round_level):
        self.value = round(self.value, round_level)
        return self

    def to_json(self, with_calculated_attributes_data=False):
        output_dict = {
            "label": self.label, "value": float(self.value.magnitude), "unit": str(self.value.units)}

        if self.source is not None:
            output_dict["source"] = {"name": self.source.name, "link": self.source.link}

        if with_calculated_attributes_data:
            output_dict["id"] = self.id
            output_dict["direct_ancestors_with_id"] = [elt.id for elt in self.direct_ancestors_with_id]
            output_dict["direct_children_with_id"] = [elt.id for elt in self.direct_children_with_id]

        return output_dict

    def __repr__(self):
        return json.dumps(self.to_json())

    def __str__(self):
        if isinstance(self.value, Quantity):
            return f"{round(self.value, 2)}"
        else:
            return str(self.value)


class ExplainableHourlyQuantities(ExplainableObject):
    def __init__(
            self, value: pd.DataFrame, label: str = None, left_parent: ExplainableObject = None,
            right_parent: ExplainableObject = None, operator: str = None, source: Source = None):
        if not isinstance(value, pd.DataFrame):
            raise ValueError(f"ExplainableHourlyQuantities values must be pandas DataFrames, got {type(value)}")
        if value.columns != ["value"]:
            raise ValueError(
                f"ExplainableHourlyQuantities values must have only one column named value, got {value.columns}")
        if not isinstance(value.dtypes.iloc[0], pint_pandas.pint_array.PintType):
            raise ValueError(f"The pd DataFrame value of an ExplainableHourlyQuantities object must be typed with Pint,"
                             f" got {type(value.dtypes.iloc[0])} dtype")
        super().__init__(value, label, left_parent, right_parent, operator, source)

    def to(self, unit_to_convert_to: Unit, rounding=None):
        self.value["value"] = self.value["value"].pint.to(unit_to_convert_to)
        if rounding is not None:
            self.value["value"] = pint_pandas.PintArray(
                [round(elt, rounding) for elt in self.value["value"].values._data], dtype=self.unit)

        return self

    def return_shifted_hourly_quantities(self, hours: int):
        return ExplainableHourlyQuantities(
            self.value.shift(hours, freq="h"), left_parent=self, operator=f"shift by {hours}")

    @property
    def unit(self):
        return self.value.dtypes.iloc[0].units

    @property
    def value_as_float_list(self):
        return [float(elt) for elt in self.value["value"].values._data]

    def convert_to_utc(self, local_timezone):
        utc_tz = pytz.timezone('UTC')
        current_time = datetime.now()
        time_diff = (utc_tz.utcoffset(current_time) - local_timezone.value.utcoffset(current_time))
        time_diff_in_hours = int(time_diff.total_seconds() / 3600)

        return ExplainableHourlyQuantities(
            self.value.copy().shift(time_diff_in_hours, freq="h"),
            left_parent=self, right_parent=local_timezone, operator="converted to UTC from")

    def sum(self):
        return ExplainableQuantity(self.value["value"].sum(), left_parent=self, operator="sum")

    def mean(self):
        return ExplainableQuantity(self.value["value"].mean() / 24, left_parent=self, operator="mean")

    def max(self):
        return ExplainableQuantity(self.value["value"].max(), left_parent=self, operator="max")

    def abs(self):
        return ExplainableHourlyQuantities(
            pd.DataFrame(
                {"value": pint_pandas.PintArray(np.abs(self.value["value"].values.data), dtype=self.unit)},
                index=self.value.index),
            left_parent=self, operator="abs")

    def ceil(self):
        return ExplainableHourlyQuantities(
            pd.DataFrame(
                {"value": pint_pandas.PintArray(np.ceil(self.value["value"].values.data), dtype=self.unit)},
                index=self.value.index),
            left_parent=self, operator="ceil")

    def copy(self):
        return ExplainableHourlyQuantities(self.value.copy(), left_parent=self, operator="duplicate")

    def __eq__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return False
        elif isinstance(other, EmptyExplainableObject):
            return False
        if issubclass(type(other), ExplainableHourlyQuantities):
            if len(self.value) != len(other.value):
                raise ValueError(
                    f"Can only compare ExplainableHourlyUsages with values of same length. Here we are trying to "
                    f"compare {self.value} and {other.value}.")

            return (self.value["value"] == other.value["value"]).all()
        else:
            raise ValueError(f"Can only compare with another ExplainableHourlyUsage, not {type(other)}")

    def __len__(self):
        return len(self.value)

    def __add__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            # summing with sum() adds an implicit 0 as starting value
            return ExplainableHourlyQuantities(self.value, left_parent=self)
        elif isinstance(other, EmptyExplainableObject):
            return ExplainableHourlyQuantities(self.value, left_parent=self)
        elif issubclass(type(other), ExplainableHourlyQuantities):
            df_sum = self.value.add(other.value, fill_value=0 * self.unit)
            return ExplainableHourlyQuantities(df_sum, "", self, other, "+")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage, not with {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return ExplainableHourlyQuantities(self.value, left_parent=self)
        elif isinstance(other, EmptyExplainableObject):
            return ExplainableHourlyQuantities(self.value, left_parent=self)
        elif issubclass(type(other), ExplainableHourlyQuantities):
            return ExplainableHourlyQuantities(self.value - other.value, "", self, other, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage, not with {type(other)}")

    def __rsub__(self, other):
        if issubclass(type(other), ExplainableHourlyQuantities):
            return ExplainableHourlyQuantities(other.value - self.value, "", other, self, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage, not with {type(other)}")

    def __mul__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return 0
        elif isinstance(other, EmptyExplainableObject):
            return EmptyExplainableObject()
        elif issubclass(type(other), ExplainableQuantity) or issubclass(type(other), ExplainableHourlyQuantities):
            return ExplainableHourlyQuantities(self.value.mul(other.value, fill_value=0), "", self, other, "*")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, "
                f"not with {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if issubclass(type(other), ExplainableHourlyQuantities):
            raise NotImplementedError
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableHourlyQuantities(self.value / other.value, "", self, other, "/")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, "
                f"not with {type(other)}")

    def __rtruediv__(self, other):
        if issubclass(type(other), ExplainableHourlyQuantities):
            raise NotImplementedError
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableHourlyQuantities(other.value / self.value, "", other, self, "/")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity,"
                f" not with {type(other)}")

    def to_json(self, with_calculated_attributes_data=False):
        output_dict = {
            "label": self.label,
            "values": list(map(lambda x: round(float(x), 2), self.value["value"].values._data)),
            "unit": str(self.value.dtypes.iloc[0].units),
            "start_date": self.value.index[0].strftime("%Y-%m-%d %H:%M:%S")
        }

        if self.source is not None:
            output_dict["source"] = {"name": self.source.name, "link": self.source.link}

        if with_calculated_attributes_data:
            output_dict["id"] = self.id
            output_dict["direct_ancestors_with_id"] = [elt.id for elt in self.direct_ancestors_with_id]
            output_dict["direct_children_with_id"] = [elt.id for elt in self.direct_children_with_id]

        return output_dict

    def __repr__(self):
        return json.dumps(self.to_json())

    def __str__(self):
        def _round_series_values(input_series):
            return [str(round(hourly_value.magnitude, 2)) for hourly_value in input_series.tolist()]

        compact_unit = "{:~}".format(self.unit)
        if self.unit == u.dimensionless:
            compact_unit = "dimensionless"

        nb_of_values = len(self.value)
        if nb_of_values < 30:
            rounded_values = _round_series_values(self.value["value"])
            str_rounded_values = "[" + ", ".join(rounded_values) + "]"
        else:
            first_vals = _round_series_values(self.value["value"].iloc[:10])
            last_vals = _round_series_values(self.value["value"].iloc[-10:])
            str_rounded_values = "first 10 vals [" + ", ".join(first_vals) \
                                 + "],\n    last 10 vals [" + ", ".join(last_vals) + "]"

        return f"{nb_of_values} values from {self.value.index.min().to_timestamp()} " \
               f"to {self.value.index.max().to_timestamp()} in {compact_unit}:\n    {str_rounded_values}"

    def plot_on_ax(self, ax):
        ax.plot(self.value.index.to_timestamp().values, self.value["value"].values.data)
        plt.ylabel(f"{self.unit:~}")

        locator = mdates.AutoDateLocator(minticks=3, maxticks=12)
        formatter = mdates.ConciseDateFormatter(locator)
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter(formatter)

    def plot(self, figsize=(10, 4), filepath=None, plt_show=False, xlims=None):
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

        if xlims is not None:
            start, end = xlims
            filtered_df = self.value[(self.value.index.to_timestamp() >= start)
                                     & (self.value.index.to_timestamp() <= end)]
            ax.set_xlim(xlims)
            max_val = filtered_df["value"].max().magnitude
            min_val = filtered_df["value"].min().magnitude
            offset = (max_val - min_val) * 0.1
            ax.set_ylim([min_val - offset, max_val + offset])

        self.plot_on_ax(ax)

        if self.label:
            ax.set_title(self.label)

        if filepath is not None:
            plt.savefig(filepath, bbox_inches='tight')

        if plt_show:
            plt.show()



from datetime import datetime, timedelta
from typing import List

import numpy as np
import pandas as pd
import pint

from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.source_objects import SourceHourlyValues


def create_random_hourly_usage_df(
        nb_days: int = 1, min_val: int = 1, max_val: int = 10,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    end_date = start_date + timedelta(days=nb_days)
    period_index = pd.period_range(start=start_date, end=end_date, freq='h')

    data = np.random.randint(min_val, max_val, size=len(period_index))
    df = pd.DataFrame(data, index=period_index, columns=['value'], dtype=f"pint[{str(pint_unit)}]")

    return df


def create_hourly_usage_df_from_list(
        input_list: List[float], start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    end_date = start_date + timedelta(hours=len(input_list) - 1)
    period_index = pd.period_range(start=start_date, end=end_date, freq='h')

    df = pd.DataFrame(input_list, index=period_index, columns=['value'], dtype=f"pint[{str(pint_unit)}]")

    return df


def linear_growth_hourly_values(
        nb_of_hours: int, start_value: int, end_value: int,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    linear_growth = np.linspace(start_value, end_value, nb_of_hours)

    df = create_hourly_usage_df_from_list(linear_growth, start_date, pint_unit)

    return SourceHourlyValues(df)


def sinusoidal_fluct_hourly_values(
        nb_of_hours: int, sin_fluct_amplitude: int, sin_fluct_period_in_hours: int,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    time = np.arange(nb_of_hours)
    sinusoidal_fluctuation = sin_fluct_amplitude * np.sin(2 * np.pi * time / sin_fluct_period_in_hours)

    df = create_hourly_usage_df_from_list(sinusoidal_fluctuation, start_date, pint_unit)

    return SourceHourlyValues(df)


def daily_fluct_hourly_values(nb_of_hours: int, fluct_scale: float, hour_of_day_for_min_value: int = 4,
                              start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
                              pint_unit: pint.Unit = u.dimensionless):
    assert fluct_scale > 0
    assert fluct_scale <= 1
    time = np.arange(nb_of_hours)
    hour_of_day = [(start_date.hour + x) % 24 for x in time]

    daily_fluctuation = (
            np.full(shape=len(hour_of_day), fill_value=1)
            + fluct_scale * np.sin(
                (3 * np.pi / 2)
                + (2 * np.pi
                    * (hour_of_day - np.full(shape=len(hour_of_day), fill_value=hour_of_day_for_min_value, dtype=int))
                    / 24
                )
            )
        )

    df = create_hourly_usage_df_from_list(daily_fluctuation, start_date, pint_unit)

    return SourceHourlyValues(df)



import json

from efootprint.abstract_modeling_classes.modeling_object import ModelingObject


def recursively_write_json_dict(output_dict, mod_obj, save_calculated_attributes=True):
    mod_obj_class = mod_obj.class_as_simple_str
    if mod_obj_class not in output_dict.keys():
        output_dict[mod_obj_class] = {}
    if mod_obj.id not in output_dict[mod_obj_class].keys():
        output_dict[mod_obj_class][mod_obj.id] = mod_obj.to_json(save_calculated_attributes)
        for key, value in mod_obj.__dict__.items():
            if issubclass(type(value), ModelingObject):
                recursively_write_json_dict(output_dict, value, save_calculated_attributes)
            elif type(value) == list and len(value) > 0 and issubclass(type(value[0]), ModelingObject):
                for mod_obj_elt in value:
                    recursively_write_json_dict(output_dict, mod_obj_elt, save_calculated_attributes)

    return output_dict


def system_to_json(input_system, save_calculated_attributes, output_filepath=None):
    output_dict = {}
    recursively_write_json_dict(output_dict, input_system, save_calculated_attributes)

    if output_filepath is not None:
        with open(output_filepath, "w") as file:
            file.write(json.dumps(output_dict, indent=4))

    return output_dict



from typing import List, Type

from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.core.usage.job import Job


class UserJourneyStep(ModelingObject):
    def __init__(self, name: str, user_time_spent: SourceValue, jobs: List[Job]):
        super().__init__(name)

        if not user_time_spent.value.check("[time]"):
            raise ValueError(
                "Variable 'user_time_spent' does not have the appropriate '[time]' dimensionality")
        self.user_time_spent = user_time_spent
        self.user_time_spent.set_label(f"Time spent on step {self.name}")
        self.jobs = jobs

    @property
    def user_journeys(self) -> List[Type["UserJourney"]]:
        return self.modeling_obj_containers

    @property
    def usage_patterns(self) -> List[Type["UsagePattern"]]:
        return list(set(sum([uj.usage_patterns for uj in self.user_journeys], start=[])))

    @property
    def services(self) -> List[Type["Service"]]:
        return [job.service for job in self.jobs]

    @property
    def systems(self) -> List:
        return list(set(sum([up.systems for up in self.usage_patterns], start=[])))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[Type["UserJourney"]]:
        if self.user_journeys:
            return self.user_journeys
        else:
            return self.jobs



import uuid
from abc import ABCMeta, abstractmethod
from typing import List, Type
from copy import copy
import os
import json
import re

from IPython.display import HTML

from efootprint.logger import logger
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, \
    ObjectLinkedToModelingObj
from efootprint.utils.graph_tools import WIDTH, HEIGHT, add_unique_id_to_mynetwork
from efootprint.utils.object_relationships_graphs import build_object_relationships_graph, \
    USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE
from efootprint.utils.tools import convert_to_list


PREVIOUS_LIST_VALUE_SET_SUFFIX = "__previous_list_value_set"


def get_subclass_attributes(obj, target_class):
    return {attr_name: attr_value for attr_name, attr_value in obj.__dict__.items()
            if issubclass(type(attr_value), target_class)}


def check_type_homogeneity_within_list_or_set(input_list_or_set):
    type_set = [type(value) for value in input_list_or_set]
    base_type = type(type_set[0])

    if not all(isinstance(item, base_type) for item in type_set):
        raise ValueError(
            f"There shouldn't be objects of different types within the same list, found {type_set}")
    else:
        return type_set.pop()


class AfterInitMeta(type):
    def __call__(cls, *args, **kwargs):
        instance = super(AfterInitMeta, cls).__call__(*args, **kwargs)
        instance.after_init()
        return instance


class ABCAfterInitMeta(ABCMeta, AfterInitMeta):
    pass


def css_escape(input_string):
    """
    Escape a string to be used as a CSS identifier.
    """
    def escape_char(c):
        if re.match(r'[a-zA-Z0-9_-]', c):
            return c
        elif c == ' ':
            return '-'
        else:
            return f'{ord(c):x}'

    return ''.join(escape_char(c) for c in input_string)


class ModelingObject(metaclass=ABCAfterInitMeta):
    def __init__(self, name):
        self.dont_handle_input_updates = False
        self.init_has_passed = False
        self.name = name
        self.id = f"id-{str(uuid.uuid4())[:6]}-{css_escape(self.name)}"
        self.modeling_obj_containers = []

    @property
    @abstractmethod
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[Type["ModelingObject"]]:
        pass

    @property
    def calculated_attributes(self) -> List[str]:
        return []

    @property
    @abstractmethod
    def systems(self) -> List:
        pass

    def compute_calculated_attributes(self):
        logger.info(f"Computing calculated attributes for {type(self).__name__} {self.name}")
        for attr_name in self.calculated_attributes:
            update_func = self.retrieve_update_function_from_attribute_name(self, attr_name)
            update_func()

    @property
    def attributes_computation_chain(self):
        attributes_computation_chain = [self]

        mod_objs_with_attributes_to_compute = self.modeling_objects_whose_attributes_depend_directly_on_me

        while len(mod_objs_with_attributes_to_compute) > 0:
            current_mod_obj_to_update = mod_objs_with_attributes_to_compute[0]
            attributes_computation_chain.append(current_mod_obj_to_update)
            mod_objs_with_attributes_to_compute = mod_objs_with_attributes_to_compute[1:]

            for mod_obj in current_mod_obj_to_update.modeling_objects_whose_attributes_depend_directly_on_me:
                if mod_obj not in mod_objs_with_attributes_to_compute:
                    mod_objs_with_attributes_to_compute.append(mod_obj)

        return attributes_computation_chain

    @staticmethod
    def optimize_attributes_computation_chain(attributes_computation_chain):
        # Keep only last occurrence of each mod_obj
        optimized_chain = []

        for index in range(len(attributes_computation_chain)):
            mod_obj = attributes_computation_chain[index]

            if mod_obj not in attributes_computation_chain[index + 1:]:
                optimized_chain.append(mod_obj)

        return optimized_chain

    @staticmethod
    def launch_attributes_computation_chain(attributes_computation_chain):
        for mod_obj in attributes_computation_chain:
            mod_obj.compute_calculated_attributes()

    def after_init(self):
        self.init_has_passed = True

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if issubclass(type(other), ModelingObject):
            return self.id == other.id
        return False

    def register_footprint_values_in_systems_before_change(self, change: str):
        logger.debug(change)
        for system in self.systems:
            system.previous_total_energy_footprints = system.total_energy_footprint_sum_over_period
            system.previous_total_fabrication_footprints = system.total_fabrication_footprint_sum_over_period
            system.previous_change = change
            system.all_changes.append(change)

    def __setattr__(self, name, input_value):
        old_value = self.__dict__.get(name, None)

        if name not in ["dont_handle_input_updates", "init_has_passed"] and not self.dont_handle_input_updates:
            if issubclass(type(input_value), ModelingObject):
                input_value.add_obj_to_modeling_obj_containers(self)
                if self.init_has_passed:
                    self.register_footprint_values_in_systems_before_change(
                        f"{self.name}’s {name} changed from {old_value.name} to {input_value.name}")
                    super().__setattr__(name, input_value)
                    self.handle_object_link_update(input_value, old_value)

            elif issubclass(type(input_value), List) and name not in ["modeling_obj_containers"]:
                if len(input_value) > 0 and type(input_value[0]) == str and self.init_has_passed:
                    raise ValueError(f"There shouldn’t be a str list update after init")
                old_list_value_attr_name = f"{name}{PREVIOUS_LIST_VALUE_SET_SUFFIX}"
                if not (len(input_value) > 0 and type(input_value[0]) == str):
                    for obj in input_value:
                        obj.add_obj_to_modeling_obj_containers(self)
                    # Necessary to handle syntax obj.list_attr += [new_attr_in_list] because lists are mutable objects
                    # Otherwise if using old_value, it would already be equal to input_value
                    old_list_value = getattr(self, old_list_value_attr_name, None)
                    if self.init_has_passed and old_list_value is not None:
                        oldlist_ids = [mod_obj.name for mod_obj in old_list_value]
                        newlist_ids = [mod_obj.name for mod_obj in input_value]
                        # Reset list to old value before registering footprints
                        super().__setattr__(name, old_list_value)
                        self.register_footprint_values_in_systems_before_change(
                            f"{self.name}’s {name} changed from {oldlist_ids} to {newlist_ids}")
                        super().__setattr__(name, input_value)
                        self.handle_object_list_link_update(input_value, old_list_value)
                    super().__setattr__(old_list_value_attr_name, copy(input_value))

            elif issubclass(type(input_value), ObjectLinkedToModelingObj):
                input_value.set_modeling_obj_container(self, name)
                is_a_user_attribute_update = self.init_has_passed and (
                    name not in self.calculated_attributes and old_value is not None)
                if is_a_user_attribute_update:
                    self.register_footprint_values_in_systems_before_change(
                        f"{self.name}’s {name} changed from {str(old_value)} to {str(input_value)}")
                    super().__setattr__(name, input_value)
                    self.handle_model_input_update(old_value)

        super().__setattr__(name, input_value)

        if getattr(self, "name", None) is not None:
            logger.debug(f"attribute {name} updated in {self.name}")

    @staticmethod
    def retrieve_update_function_from_attribute_name(mod_obj, attr_name):
        update_func_name = f"update_{attr_name}"
        update_func = getattr(mod_obj, update_func_name, None)

        if update_func is None:
            raise AttributeError(f"No update function associated to {attr_name} in {mod_obj.id}. "
                                 f"Please create it and checkout optimization.md")

        return update_func

    def handle_model_input_update(self, old_value_that_gets_updated: ExplainableObject):
        descendants = old_value_that_gets_updated.get_all_descendants_with_id()
        has_been_recomputed_dict = {descendant.id: False for descendant in descendants}
        has_been_recomputed_dict[old_value_that_gets_updated.id] = True

        computed_parents_with_children_to_recompute = [old_value_that_gets_updated]

        while len(computed_parents_with_children_to_recompute) > 0:
            for recomputed_parent in computed_parents_with_children_to_recompute:
                drop_recomputed_parent_from_list = True
                for child in recomputed_parent.direct_children_with_id:
                    if not has_been_recomputed_dict[child.id]:
                        ancestors_that_belong_to_old_value_descendants = [
                            ancestor for ancestor in child.direct_ancestors_with_id
                            if ancestor.id in [ancestor.id for ancestor in descendants]]
                        if all([has_been_recomputed_dict[ancestor.id]
                                for ancestor in ancestors_that_belong_to_old_value_descendants]):
                            child_update_func = self.retrieve_update_function_from_attribute_name(
                                child.modeling_obj_container, child.attr_name_in_mod_obj_container)
                            child_update_func()
                            has_been_recomputed_dict[child.id] = True
                            if len(child.direct_children_with_id) > 0:
                                computed_parents_with_children_to_recompute.append(child)
                        else:
                            # Wait for next iteration
                            drop_recomputed_parent_from_list = False
                if drop_recomputed_parent_from_list:
                    computed_parents_with_children_to_recompute = [
                        child for child in computed_parents_with_children_to_recompute
                        if child.id != recomputed_parent.id]

    def handle_object_link_update(
            self, input_value: Type["ModelingObject"], old_value: Type["ModelingObject"]):
        if old_value is None:
            raise ValueError(f"A link update is trying to replace an null object")
        if (self in old_value.modeling_objects_whose_attributes_depend_directly_on_me and
            old_value in self.modeling_objects_whose_attributes_depend_directly_on_me):
            raise AssertionError(
                f"There is a circular recalculation dependency between {self.id} and {old_value.id}")

        old_value.remove_obj_from_modeling_obj_containers(self)

        attributes_computation_chain = []
        if self in old_value.modeling_objects_whose_attributes_depend_directly_on_me:
            attributes_computation_chain += self.attributes_computation_chain
        else:
            attributes_computation_chain += input_value.attributes_computation_chain
            attributes_computation_chain += old_value.attributes_computation_chain

        optimized_chain = self.optimize_attributes_computation_chain(attributes_computation_chain)
        self.launch_attributes_computation_chain(optimized_chain)

    def handle_object_list_link_update(self, input_value: List[Type["ModelingObject"]],
                                       old_value: List[Type["ModelingObject"]]):
        removed_objs = [obj for obj in old_value if obj not in input_value]
        added_objs = [obj for obj in input_value if obj not in old_value]

        for obj in removed_objs:
            obj.remove_obj_from_modeling_obj_containers(self)

        attributes_computation_chain = []

        for obj in removed_objs + added_objs:
            if self not in obj.modeling_objects_whose_attributes_depend_directly_on_me:
                attributes_computation_chain += obj.attributes_computation_chain

        attributes_computation_chain += self.attributes_computation_chain

        optimized_chain = self.optimize_attributes_computation_chain(attributes_computation_chain)
        self.launch_attributes_computation_chain(optimized_chain)

    def add_obj_to_modeling_obj_containers(self, new_obj):
        if new_obj not in self.modeling_obj_containers:
            if (len(self.modeling_obj_containers) > 0
                    and not isinstance(new_obj, type(self.modeling_obj_containers[0]))):
                raise ValueError(
                    f"There shouldn't be objects of different types within modeling_obj_containers for {self.name},"
                    f" found {type(new_obj)} and {type(self.modeling_obj_containers[0])}")
            self.modeling_obj_containers.append(new_obj)

    def remove_obj_from_modeling_obj_containers(self, obj_to_remove):
        self.modeling_obj_containers = [
            mod_obj for mod_obj in self.modeling_obj_containers if mod_obj != obj_to_remove]

    @property
    def mod_obj_attributes(self):
        output_list = []
        for attr_name, attr_value in vars(self).items():
            values = convert_to_list(attr_value)
            for value in values:
                if isinstance(value, ModelingObject):
                    output_list.append(value)

        return output_list

    def object_relationship_graph_to_file(
            self, filename=None, classes_to_ignore=USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE, width=WIDTH, height=HEIGHT,
            notebook=False):
        object_relationships_graph = build_object_relationships_graph(
            self, classes_to_ignore=classes_to_ignore, width=width, height=height, notebook=notebook)

        if filename is None:
            filename = os.path.join(".", f"{self.name} object relationship graph.html")
        object_relationships_graph.show(filename, notebook=notebook)

        add_unique_id_to_mynetwork(filename)

        if notebook:
            return HTML(filename)

    def self_delete(self):
        logger.warning(
            f"Deleting {self.name}, removing backward links pointing to it in "
            f"{','.join([mod_obj.name for mod_obj in self.mod_obj_attributes])}")
        if self.modeling_obj_containers:
            raise PermissionError(
                f"You can’t delete {self.name} because "
                f"{','.join([mod_obj.name for mod_obj in self.modeling_obj_containers])} have it as attribute.")

        attributes_computation_chain = []
        for attr in self.mod_obj_attributes:
            attr.modeling_obj_containers = [elt for elt in attr.modeling_obj_containers if elt != self]
            attributes_computation_chain += attr.attributes_computation_chain

        optimized_chain = self.optimize_attributes_computation_chain(attributes_computation_chain)
        self.launch_attributes_computation_chain(optimized_chain)

        del self

    def to_json(self, save_calculated_attributes=False):
        output_dict = {}

        for key, value in self.__dict__.items():
            if (
                    (key in self.calculated_attributes and not save_calculated_attributes)
                    or key in ["all_changes", "modeling_obj_containers"]
                    or key.startswith("previous")
                    or key.startswith("initial")
            ):
                continue
            if type(value) == str:
                output_dict[key] = value
            elif type(value) == int:
                output_dict[key] = value
            elif type(value) == list and PREVIOUS_LIST_VALUE_SET_SUFFIX not in key:
                if len(value) == 0:
                    output_dict[key] = value
                else:
                    if type(value[0]) == str:
                        output_dict[key] = value
                    elif issubclass(type(value[0]), ModelingObject):
                        output_dict[key] = [elt.id for elt in value]
            elif issubclass(type(value), ExplainableObject):
                output_dict[key] = value.to_json(save_calculated_attributes)
            elif issubclass(type(value), ModelingObject):
                output_dict[key] = value.id

        return output_dict

    @property
    def class_as_simple_str(self):
        return str(self.__class__).replace("<class '", "").replace("'>", "").split(".")[-1]

    def __repr__(self):
        return json.dumps(self.to_json(save_calculated_attributes=True), indent=4)

    def __str__(self):
        output_str = ""

        def key_value_to_str(input_key, input_value):
            key_value_str = ""

            if type(input_value) in (str, int) or input_value is None:
                key_value_str = f"{input_key}: {input_value}\n"
            elif type(input_value) == list:
                if len(input_value) == 0:
                    key_value_str = f"{input_key}: {input_value}\n"
                else:
                    if type(input_value[0]) == str:
                        key_value_str = f"{input_key}: {input_value}"
                    elif issubclass(type(input_value[0]), ModelingObject) and PREVIOUS_LIST_VALUE_SET_SUFFIX not in key:
                        str_value = "[" + ", ".join([elt.id for elt in input_value]) + "]"
                        key_value_str = f"{input_key}: {str_value}\n"
            elif issubclass(type(input_value), ExplainableObject):
                key_value_str = f"{input_key}: {input_value}\n"
            elif issubclass(type(input_value), ModelingObject):
                key_value_str = f"{input_key}: {input_value.id}\n"

            return key_value_str

        output_str += f"{self.class_as_simple_str} {self.id}\n \n"

        for key, attr_value in self.__dict__.items():
            if key == "modeling_obj_containers" or key in self.calculated_attributes or key.startswith("previous")\
                    or key in ["name", "id"]:
                continue
            output_str += key_value_to_str(key, attr_value)

        if len(self.calculated_attributes) > 0:
            output_str += " \ncalculated_attributes:\n"
            for key in self.calculated_attributes:
                output_str += "  " + key_value_to_str(key, getattr(self, key))

        return output_str



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.storage import Storage


def default_ssd(name="Default SSD storage", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(160 * u.kg, Sources.STORAGE_EMBODIED_CARBON_STUDY),
        "power": SourceValue(1.3 * u.W, Sources.STORAGE_EMBODIED_CARBON_STUDY),
        "lifespan": SourceValue(6 * u.years, Sources.HYPOTHESIS),
        "idle_power": SourceValue(0 * u.W, Sources.HYPOTHESIS),
        "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
        "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
        "average_carbon_intensity": SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
        "data_replication_factor": SourceValue(3 * u.dimensionless, Sources.HYPOTHESIS),
        "base_storage_need": SourceValue(0 * u.TB, Sources.HYPOTHESIS),
        "data_storage_duration": SourceValue(5 * u.year, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return Storage(name, **output_args)


def default_hdd(name="Default HDD storage", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(20 * u.kg, Sources.STORAGE_EMBODIED_CARBON_STUDY),
        "power": SourceValue(4.2 * u.W, Sources.STORAGE_EMBODIED_CARBON_STUDY),
        "lifespan": SourceValue(4 * u.years, Sources.HYPOTHESIS),
        "idle_power": SourceValue(0 * u.W, Sources.HYPOTHESIS),
        "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
        "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
        "average_carbon_intensity": SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
        "data_replication_factor": SourceValue(3 * u.dimensionless, Sources.HYPOTHESIS),
        "base_storage_need": SourceValue(0 * u.TB, Sources.HYPOTHESIS),
        "data_storage_duration": SourceValue(5 * u.year, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return Storage(name, **output_args)



from typing import Dict, List

import plotly.express as px
import plotly
import pandas as pd
from IPython.display import HTML
from matplotlib import pyplot as plt

from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.units import u
from efootprint.core.hardware.network import Network
from efootprint.core.hardware.servers.server_base_class import Server
from efootprint.core.hardware.storage import Storage
from efootprint.core.service import Service
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.usage.user_journey import UserJourney
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, EmptyExplainableObject, \
    ExplainableHourlyQuantities
from efootprint.logger import logger
from efootprint.utils.plot_emission_diffs import EmissionPlotter
from efootprint.utils.tools import format_co2_amount, display_co2_amount


class System(ModelingObject):
    def __init__(self, name: str, usage_patterns: List[UsagePattern]):
        super().__init__(name)
        self.usage_patterns = usage_patterns
        self.previous_change = None
        self.previous_total_energy_footprints = None
        self.previous_total_fabrication_footprints = None
        self.all_changes = []
        self.initial_total_energy_footprints = None
        self.initial_total_fabrication_footprints = None

    @property
    def user_journeys(self) -> List[UserJourney]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update({usage_pattern.user_journey})

        return list(output_set)

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self):
        return self.user_journeys

    @property
    def systems(self) -> List:
        return []

    def after_init(self):
        self.init_has_passed = True
        self.launch_attributes_computation_chain(self.attributes_computation_chain)
        logger.info(f"Finished computing {self.name} modeling")
        self.initial_total_energy_footprints = self.total_energy_footprint_sum_over_period
        self.initial_total_fabrication_footprints = self.total_fabrication_footprint_sum_over_period

    @property
    def servers(self) -> List[Server]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update(usage_pattern.user_journey.servers)

        return list(output_set)

    @property
    def storages(self) -> List[Storage]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update(usage_pattern.user_journey.storages)

        return list(output_set)

    @property
    def services(self) -> List[Service]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update(usage_pattern.user_journey.services)

        return list(output_set)

    @property
    def networks(self) -> List[Network]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update({usage_pattern.network})

        return list(output_set)

    def get_storage_by_name(self, storage_name) -> Storage:
        for storage in self.storages:
            if storage.name == storage_name:
                return storage

    def get_server_by_name(self, server_name) -> Server:
        for server in self.servers:
            if server.name == server_name:
                return server

    def get_usage_pattern_by_name(self, usage_pattern_name) -> UsagePattern:
        for usage_pattern in self.usage_patterns:
            if usage_pattern.name == usage_pattern_name:
                return usage_pattern

    @property
    def fabrication_footprints(self) -> Dict[str, Dict[str, ExplainableHourlyQuantities]]:
        fab_footprints = {
            "Servers": {server.name: server.instances_fabrication_footprint for server in self.servers},
            "Storage": {storage.name: storage.instances_fabrication_footprint for storage in self.storages},
            "Network": {"networks": EmptyExplainableObject()},
            "Devices": {usage_pattern.name: usage_pattern.devices_fabrication_footprint
                        for usage_pattern in self.usage_patterns},
        }

        return fab_footprints

    @property
    def energy_footprints(self) -> Dict[str, Dict[str, ExplainableHourlyQuantities]]:
        energy_footprints = {
            "Servers": {server.name: server.energy_footprint for server in self.servers},
            "Storage": {storage.name: storage.energy_footprint for storage in self.storages},
            "Network": {network.name: network.energy_footprint for network in self.networks},
            "Devices": {usage_pattern.name: usage_pattern.devices_energy_footprint
                        for usage_pattern in self.usage_patterns},
        }

        return energy_footprints

    @property
    def total_fabrication_footprints(self) -> Dict[str, ExplainableHourlyQuantities]:
        fab_footprints = {
            "Servers": sum(server.instances_fabrication_footprint for server in self.servers),
            "Storage": sum(storage.instances_fabrication_footprint for storage in self.storages),
            "Network": EmptyExplainableObject(),
            "Devices": sum(usage_pattern.devices_fabrication_footprint
                           for usage_pattern in self.usage_patterns)
        }

        return fab_footprints

    @property
    def total_energy_footprints(self) -> Dict[str, ExplainableHourlyQuantities]:
        energy_footprints = {
            "Servers": sum(server.energy_footprint for server in self.servers),
            "Storage": sum(storage.energy_footprint for storage in self.storages),
            "Network": sum(network.energy_footprint for network in self.networks),
            "Devices": sum(usage_pattern.devices_energy_footprint for usage_pattern in self.usage_patterns)
        }

        return energy_footprints

    @staticmethod
    def sum_and_remove_empty_explainable_object(expl_obj):
        tmp_sum = expl_obj.sum()
        if isinstance(tmp_sum, EmptyExplainableObject):
            tmp_sum = ExplainableQuantity(0 * u.kg, "null value")

        return tmp_sum

    @property
    def fabrication_footprint_sum_over_period(self) -> Dict[str, Dict[ModelingObject, ExplainableQuantity]]:
        fab_footprints_sum = {}
        for key, dict_value in self.fabrication_footprints.items():
            fab_footprints_sum[key] = {
                obj_key: self.sum_and_remove_empty_explainable_object(obj_value)
                for obj_key, obj_value in dict_value.items()
            }

        return fab_footprints_sum

    @property
    def energy_footprint_sum_over_period(self) -> Dict[str, Dict[ModelingObject, ExplainableQuantity]]:
        energy_footprints_sum = {}
        for key, dict_value in self.energy_footprints.items():
            energy_footprints_sum[key] = {
                obj_key: self.sum_and_remove_empty_explainable_object(obj_value)
                for obj_key, obj_value in dict_value.items()
            }

        return energy_footprints_sum

    @property
    def total_fabrication_footprint_sum_over_period(self) -> Dict[str, ExplainableQuantity]:
        fab_footprints = {
            obj_key: self.sum_and_remove_empty_explainable_object(obj_value)
            for obj_key, obj_value in self.total_fabrication_footprints.items()
        }
        return fab_footprints

    @property
    def total_energy_footprint_sum_over_period(self) -> Dict[str, ExplainableQuantity]:
        energy_footprints = {
            obj_key: self.sum_and_remove_empty_explainable_object(obj_value)
            for obj_key, obj_value in self.total_energy_footprints.items()
        }
        return energy_footprints

    @property
    def total_footprint(self):
        return (
            sum(
                sum(
                    self.fabrication_footprints[key].values()) + sum(self.energy_footprints[key].values())
                for key in self.fabrication_footprints.keys()
            )
        ).set_label(f"{self.name} total carbon footprint")

    def plot_footprints_by_category_and_object(self, filename=None, height=400, width=800, return_only_html=False):
        fab_footprints = self.fabrication_footprint_sum_over_period
        energy_footprints = self.energy_footprint_sum_over_period
        categories = list(fab_footprints.keys())

        rows_as_dicts = []

        value_colname = "tonnes CO2 emissions"
        for category in categories:
            fab_objects = sorted(fab_footprints[category].items(), key=lambda x: x[0])
            energy_objects = sorted(energy_footprints[category].items(), key=lambda x: x[0])

            for objs, color in zip([energy_objects, fab_objects], ["Electricity", "Fabrication"]):
                data_dicts = [
                    {"Type": color, "Category": category, "Object": obj[0],
                     value_colname: obj[1].value.magnitude / 1000,
                     "Amount": f"{display_co2_amount(format_co2_amount(obj[1].value.magnitude))}"}
                    for obj in objs]
                rows_as_dicts += data_dicts

        df = pd.DataFrame.from_records(rows_as_dicts)

        total_co2 = df[value_colname].sum()

        total_footprint = self.total_footprint

        fig = px.bar(
            df, x="Category", y=value_colname, color='Type', barmode='group', height=height, width=width,
            hover_data={"Type": False, "Category": False, "Object": True, value_colname: False, "Amount": True},
            template="plotly_white",
            title=f"Total CO2 emissions from "
                  f"{total_footprint.value.index.min().to_timestamp().date()} "
                  f"to {total_footprint.value.index.max().to_timestamp().date()}"
                  f": {display_co2_amount(format_co2_amount(total_co2 * 1000, rounding_value=0))}"
                  )

        if (max(sum(energy_footprints["Servers"].values()), sum(fab_footprints["Servers"].values())) >
                max(sum(energy_footprints["Devices"].values()), sum(fab_footprints["Devices"].values()))):
            legend_alignment = "right"
            legend_x = 0.98
        else:
            legend_alignment = "left"
            legend_x = 0.02

        fig.update_layout(
            legend={"orientation": "v", "yanchor": "top", "y": 1.02, "xanchor": legend_alignment, "x": legend_x,
                    "title": ""},
            title={"x": 0.5, "y": 0.9, "xanchor": 'center', "yanchor": 'top'})

        total_co2_per_category_and_type = df.groupby(["Category", "Type"])[value_colname].sum()

        for category, source_type in total_co2_per_category_and_type.keys():
            height = total_co2_per_category_and_type.loc[category, source_type]
            x_shift_direction = 1 if source_type == 'Fabrication' else -1

            fig.add_annotation(
                x=category,
                y=height,
                text=f"{int((height / total_co2) * 100)}%",  # Format the label as a percentage
                showarrow=False,
                yshift=10,  # Shift the label slightly above the stack
                xshift=30 * x_shift_direction
            )

        if return_only_html:
            return fig.to_html(full_html=False, include_plotlyjs=False)

        else:
            if filename is None:
                filename = f"{self.name} footprints.html"

            plotly.offline.plot(fig, filename=filename, auto_open=False)

            return HTML(filename)

    def plot_emission_diffs(self, filepath=None, figsize=(10, 5), from_start=False, plt_show=False):
        if self.previous_change is None:
            raise ValueError(
                f"There has been no change to the system yet so no diff to plot.\n"
                f"Use System.plot_footprints_by_category_and_object() to visualize footprints")

        if from_start and len(self.all_changes) > 1:
            changes_list = "\n- ".join([change.replace('changed', 'changing') for change in self.all_changes])
            print(f"Plotting the impact of:\n\n- {changes_list}")
            emissions_dict__old = [self.initial_total_energy_footprints, self.initial_total_fabrication_footprints]
        else:
            print(f"Plotting the impact of {self.previous_change.replace('changed', 'changing')}")
            emissions_dict__old = [self.previous_total_energy_footprints, self.previous_total_fabrication_footprints]

        emissions_dict__new = [self.total_energy_footprint_sum_over_period, self.total_fabrication_footprint_sum_over_period]

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

        EmissionPlotter(
            ax, emissions_dict__old, emissions_dict__new, rounding_value=0).plot_emission_diffs()

        if filepath is not None:
            plt.savefig(filepath, bbox_inches='tight')

        if plt_show:
            plt.show()



from typing import List

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.constants.countries import Country
from efootprint.constants.units import u
from efootprint.core.hardware.hardware_base_classes import Hardware
from efootprint.core.usage.user_journey import UserJourney
from efootprint.core.usage.compute_nb_occurrences_in_parallel import compute_nb_avg_hourly_occurrences
from efootprint.core.service import Service
from efootprint.core.usage.job import Job
from efootprint.core.hardware.network import Network
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_objects import (
    ExplainableQuantity, ExplainableHourlyQuantities, EmptyExplainableObject)


class UsagePattern(ModelingObject):
    def __init__(self, name: str, user_journey: UserJourney, devices: List[Hardware], network: Network,
                 country: Country, hourly_user_journey_starts: ExplainableHourlyQuantities):
        super().__init__(name)
        self.utc_hourly_user_journey_starts = None
        self.nb_user_journeys_in_parallel = None
        self.devices_energy = None
        self.devices_energy_footprint = None
        self.devices_fabrication_footprint = None
        self.energy_footprint = None
        self.instances_fabrication_footprint = None
        self.hourly_job_occurrences_per_job = ExplainableObjectDict()
        self.hourly_avg_job_occurrences_per_job = ExplainableObjectDict()
        self.hourly_data_upload_per_job = ExplainableObjectDict()
        self.hourly_data_download_per_job = ExplainableObjectDict()
        self.hourly_user_journey_starts = hourly_user_journey_starts.set_label(f"{self.name} hourly nb of visits")
        self.user_journey = user_journey
        self.devices = devices
        self.network = network
        self.country = country

    @property
    def calculated_attributes(self):
        return ["utc_hourly_user_journey_starts", "nb_user_journeys_in_parallel", "devices_energy",
                "devices_energy_footprint", "devices_fabrication_footprint", "energy_footprint",
                "instances_fabrication_footprint"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return self.jobs

    @property
    def services(self) -> List[Service]:
        return self.user_journey.services

    @property
    def jobs(self) -> List[Job]:
        return list(set(sum([service.jobs for service in self.services], start=[])))

    @property
    def systems(self) -> List:
        return self.modeling_obj_containers

    def update_utc_hourly_user_journey_starts(self):
        utc_hourly_user_journey_starts = self.hourly_user_journey_starts.convert_to_utc(
            local_timezone=self.country.timezone)

        self.utc_hourly_user_journey_starts = utc_hourly_user_journey_starts.set_label(f"{self.name} UTC")

    def update_nb_user_journeys_in_parallel(self):
        nb_of_user_journeys_in_parallel = compute_nb_avg_hourly_occurrences(
            self.utc_hourly_user_journey_starts, self.user_journey.duration)

        self.nb_user_journeys_in_parallel = nb_of_user_journeys_in_parallel.set_label(
            f"{self.name} hourly nb of user journeys in parallel")

    def update_devices_energy(self):
        total_devices_energy_spent_over_one_full_hour = sum(
            [device.power for device in self.devices]) * ExplainableQuantity(1 * u.hour, "one full hour")

        devices_energy = (self.nb_user_journeys_in_parallel * total_devices_energy_spent_over_one_full_hour).to(u.kWh)

        self.devices_energy = devices_energy.set_label(f"Energy consumed by {self.name} devices")

    def update_devices_energy_footprint(self):
        energy_footprint = (self.devices_energy * self.country.average_carbon_intensity).to(u.kg)
        
        self.devices_energy_footprint = energy_footprint.set_label(f"Energy footprint of {self.name}")

    def update_devices_fabrication_footprint(self):
        devices_fabrication_footprint_over_one_hour = EmptyExplainableObject()
        for device in self.devices:
            device_uj_fabrication_footprint = (
                    device.carbon_footprint_fabrication * ExplainableQuantity(1 * u.hour, "one hour")
                    / (device.lifespan * device.fraction_of_usage_time)
            ).to(u.g).set_label(
                f"{device.name} fabrication footprint over one hour")
            devices_fabrication_footprint_over_one_hour += device_uj_fabrication_footprint

        devices_fabrication_footprint = (
                self.nb_user_journeys_in_parallel * devices_fabrication_footprint_over_one_hour).to(u.kg, rounding=2)

        self.devices_fabrication_footprint = devices_fabrication_footprint.set_label(
            f"Devices fabrication footprint of {self.name}")

    def update_energy_footprint(self):
        self.energy_footprint = (self.devices_energy_footprint + 0).set_label(f"{self.name} total energy footprint")

    def update_instances_fabrication_footprint(self):
        self.instances_fabrication_footprint = (self.devices_fabrication_footprint + 0).set_label(
            f"{self.name} total fabrication footprint")



from typing import List, Type

from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.service import Service
from efootprint.core.hardware.servers.server_base_class import Server
from efootprint.core.hardware.storage import Storage
from efootprint.core.usage.user_journey_step import UserJourneyStep
from efootprint.core.usage.job import Job


class UserJourney(ModelingObject):
    def __init__(self, name: str, uj_steps: List[UserJourneyStep]):
        super().__init__(name)
        self.duration = None
        self.uj_steps = uj_steps

    @property
    def calculated_attributes(self):
        return ["duration"]

    @property
    def servers(self) -> List[Server]:
        servers = set()
        for job in self.jobs:
            servers = servers | {job.service.server}

        return list(servers)

    @property
    def storages(self) -> List[Storage]:
        storages = set()
        for job in self.jobs:
            storages = storages | {job.service.storage}

        return list(storages)

    @property
    def services(self) -> List[Service]:
        services = set()
        for job in self.jobs:
            services = services | {job.service}

        return list(services)

    @property
    def usage_patterns(self):
        return self.modeling_obj_containers

    @property
    def systems(self) -> List:
        return list(set(sum([up.systems for up in self.usage_patterns], start=[])))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[Type["UsagePattern"]]:
        return self.usage_patterns

    @property
    def jobs(self) -> List[Job]:
        output_list = []
        for uj_step in self.uj_steps:
            output_list += uj_step.jobs

        return output_list

    def add_step(self, step: UserJourneyStep) -> None:
        step.add_obj_to_modeling_obj_containers(self)
        self.uj_steps = self.uj_steps + [step]

    def update_duration(self):
        user_time_spent_sum = sum([uj_step.user_time_spent for uj_step in self.uj_steps])

        self.duration = user_time_spent_sum.set_label(f"Duration of {self.name}")



import numpy as np
import pandas as pd
import pint_pandas

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyQuantities
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.servers.server_base_class import Server


class OnPremise(Server):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, ram: SourceValue, cpu_cores: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 server_utilization_rate: SourceValue, fixed_nb_of_instances: SourceValue = None):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, idle_power, ram, cpu_cores, power_usage_effectiveness,
            average_carbon_intensity, server_utilization_rate)
        self.fixed_nb_of_instances = None
        if fixed_nb_of_instances:
            if not fixed_nb_of_instances.value.check("[]"):
                raise ValueError("Variable 'fixed_nb_of_instances' shouldn’t have any dimensionality")
            self.fixed_nb_of_instances = fixed_nb_of_instances.set_label(
                f"User defined number of {self.name} instances").to(u.dimensionless)

    def update_nb_of_instances(self):
        max_nb_of_instances = self.raw_nb_of_instances.max().ceil().to(u.dimensionless)

        nb_of_instances_df = pd.DataFrame(
            {"value": pint_pandas.PintArray(
                    max_nb_of_instances.magnitude * np.ones(len(self.raw_nb_of_instances)), dtype=u.dimensionless)},
            index=self.raw_nb_of_instances.value.index
        )

        nb_of_instances = ExplainableHourlyQuantities(
            nb_of_instances_df,
            "Nb of instances",
            left_parent=self.raw_nb_of_instances,
            right_parent=None
        )

        if self.fixed_nb_of_instances:
            if max_nb_of_instances.value > self.fixed_nb_of_instances.value:
                raise ValueError(
                    f"The number of {self.name} instances computed from its resources need is superior to the number of"
                    f" instances specified by the user ({max_nb_of_instances.value} > {self.fixed_nb_of_instances})")
            else:
                fixed_nb_of_instances_df = pd.DataFrame(
                    {"value": pint_pandas.PintArray(
                        np.full(len(self.raw_nb_of_instances),self.fixed_nb_of_instances.value),dtype=u.dimensionless
                    )},
                    index=self.raw_nb_of_instances.value.index
                )
                nb_of_instances_re_calculate = ExplainableHourlyQuantities(
                    fixed_nb_of_instances_df,
                    "Nb of instances",
                    left_parent=self.raw_nb_of_instances,
                    right_parent=self.fixed_nb_of_instances
                )
                self.nb_of_instances = nb_of_instances_re_calculate.set_label(f"Fixed number of {self.name} instances")
        else:
            self.nb_of_instances = nb_of_instances.set_label(f"Hourly number of {self.name} instances")



from typing import List

from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.core.hardware.storage import Storage
from efootprint.core.hardware.servers.server_base_class import Server
from efootprint.constants.units import u


class Service(ModelingObject):
    def __init__(self, name: str, server: Server, storage: Storage, base_ram_consumption: SourceValue,
                 base_cpu_consumption: SourceValue = None):
        super().__init__(name)
        self.storage_needed = None
        self.hour_by_hour_cpu_need = None
        self.hour_by_hour_ram_need = None
        self.hourly_job_occurrences_across_usage_patterns_per_job = ExplainableObjectDict()
        self.hourly_avg_job_occurrences_across_usage_patterns_per_job = ExplainableObjectDict()
        self.hourly_data_upload_across_usage_patterns_per_job = ExplainableObjectDict()
        self.server = server
        self.storage = storage
        if not base_ram_consumption.value.check("[]"):
            raise ValueError("variable 'base_ram_consumption' does not have byte dimensionality")
        if base_cpu_consumption is None:
            base_cpu_consumption = SourceValue(1 * u.core)
        if not base_cpu_consumption.value.check("[cpu]"):
            raise ValueError("variable 'base_cpu_consumption' does not have core dimensionality")
        self.base_ram_consumption = base_ram_consumption.set_label(f"Base RAM consumption of {self.name}")
        self.base_cpu_consumption = base_cpu_consumption.set_label(f"Base CPU consumption of {self.name}")

    @property
    def resources_unit_dict(self):
        return {"ram": "GB", "cpu": "core"}

    @property
    def calculated_attributes(self):
        return [
            "hour_by_hour_ram_need", "hour_by_hour_cpu_need", "storage_needed"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return [self.server, self.storage]

    @property
    def jobs(self):
        return self.modeling_obj_containers

    @property
    def usage_patterns(self):
        return list(set(sum([job.usage_patterns for job in self.jobs], start=[])))

    @property
    def systems(self) -> List:
        return list(set(sum([up.systems for up in self.usage_patterns], start=[])))

    def compute_hour_by_hour_resource_need(self, resource):
        resource_unit = u(self.resources_unit_dict[resource])
        hour_by_hour_resource_needs = EmptyExplainableObject()
        for job in self.jobs:
            hour_by_hour_resource_needs += (
                job.hourly_avg_occurrences_across_usage_patterns * getattr(job, f"{resource}_needed"))

        return hour_by_hour_resource_needs.to(resource_unit).set_label(f"{self.name} hour by hour {resource} need")

    def update_hour_by_hour_ram_need(self):
        self.hour_by_hour_ram_need = self.compute_hour_by_hour_resource_need("ram")

    def update_hour_by_hour_cpu_need(self):
        self.hour_by_hour_cpu_need = self.compute_hour_by_hour_resource_need("cpu")
            
    def update_storage_needed(self):
        storage_needed = EmptyExplainableObject()

        for job in self.jobs:
            storage_needed += job.hourly_data_upload_across_usage_patterns

        self.storage_needed = storage_needed.to(u.TB).set_label(f"Hourly {self.name} storage need")



import math

from efootprint.builders.time_builders import create_hourly_usage_df_from_list
from efootprint.core.hardware.hardware_base_classes import InfraHardware
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyQuantities, \
    EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


class Storage(InfraHardware):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, storage_capacity: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 data_replication_factor: SourceValue, data_storage_duration: SourceValue,
                 base_storage_need: SourceValue):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, average_carbon_intensity)
        self.all_services_storage_needs = None
        self.storage_dumps = None
        self.storage_delta = None
        self.full_cumulative_storage_need = None
        self.long_term_storage_required = None
        self.nb_of_active_instances = None
        self.instances_power = None
        if not idle_power.value.check("[power]"):
            raise ValueError("Value of variable 'idle_power' does not have appropriate power dimensionality")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        if not storage_capacity.value.check("[]"):
            raise ValueError("Value of variable 'storage_capacity' does not have appropriate [] dimensionality")
        self.storage_capacity = storage_capacity.set_label(f"Storage capacity of {self.name}")
        if not power_usage_effectiveness.value.check("[]"):
            raise ValueError(
                "Value of variable 'power_usage_effectiveness' does not have appropriate [] dimensionality")
        self.power_usage_effectiveness = power_usage_effectiveness.set_label(f"PUE of {self.name}")
        if not data_replication_factor.value.check("[]"):
            raise ValueError("Value of variable 'data_replication_factor' does not have appropriate [] dimensionality")
        self.data_replication_factor = data_replication_factor.set_label(f"Data replication factor of {self.name}")
        if not data_storage_duration.value.check("[time]"):
            raise ValueError("Value of variable 'data_storage_duration' does not have appropriate time dimensionality")
        self.data_storage_duration = data_storage_duration.set_label(f"Data storage duration of {self.name}")
        if not base_storage_need.value.check("[]"):
            raise ValueError(
                "Value of variable 'storage_need_from_previous_year' does not have the appropriate"
                " '[]' dimensionality")
        self.base_storage_need = base_storage_need.set_label(f"{self.name} initial storage need")

    @property
    def calculated_attributes(self):
        return [
            "all_services_storage_needs", "storage_dumps", "storage_delta", "full_cumulative_storage_need",
            "nb_of_active_instances"] + self.calculated_attributes_defined_in_infra_hardware_class

    def update_all_services_storage_needs(self):
        all_services_storage_needs = (
                sum([service.storage_needed for service in self.services], start=EmptyExplainableObject()).to(u.TB)
                * self.data_replication_factor)

        self.all_services_storage_needs = all_services_storage_needs.set_label(
            f"Storage need of {self.name}")

    def update_storage_dumps(self):
        if isinstance(self.all_services_storage_needs, EmptyExplainableObject):
            self.storage_dumps = EmptyExplainableObject()
        else:
            storage_duration_in_hours = math.ceil(self.data_storage_duration.to(u.hour).magnitude)
            storage_dumps_df = - self.all_services_storage_needs.value.copy().shift(
                periods=storage_duration_in_hours, freq='h')
            storage_dumps_df = storage_dumps_df[
                storage_dumps_df.index <= self.all_services_storage_needs.value.index.max()]

            if len(storage_dumps_df) == 0:
                storage_needs_start_date = self.all_services_storage_needs.value.index.min().to_timestamp()
                storage_needs_end_date = self.all_services_storage_needs.value.index.max().to_timestamp()
                storage_needs_nb_of_hours = int((storage_needs_end_date - storage_needs_start_date).seconds / 3600)
                storage_dumps_df = create_hourly_usage_df_from_list(
                    [0] * (storage_needs_nb_of_hours + 1), start_date=storage_needs_start_date)

            self.storage_dumps = ExplainableHourlyQuantities(
                storage_dumps_df, label=f"Storage dumps for {self.name}",
                left_parent=self.all_services_storage_needs,
                right_parent=self.data_storage_duration, operator="shift by storage duration and negate")

    def update_storage_delta(self):
        storage_delta = self.all_services_storage_needs + self.storage_dumps
        
        self.storage_delta = storage_delta.set_label(f"Hourly storage delta for {self.name}")

    def update_full_cumulative_storage_need(self):
        if isinstance(self.storage_delta, EmptyExplainableObject):
            self.full_cumulative_storage_need = EmptyExplainableObject()
        else:
            storage_delta_df = self.storage_delta.value.copy()
            storage_delta_df.iat[0, 0] += self.base_storage_need.value
            full_cumulative_storage_need = storage_delta_df.cumsum()

            self.full_cumulative_storage_need = ExplainableHourlyQuantities(
                full_cumulative_storage_need, label=f"Full cumulative storage need for {self.name}",
                left_parent=self.storage_delta, right_parent=self.base_storage_need,
                operator="cumulative sum of storage delta with initial storage need")

    def update_raw_nb_of_instances(self):
        raw_nb_of_instances = (self.full_cumulative_storage_need / self.storage_capacity).to(u.dimensionless)

        self.raw_nb_of_instances = raw_nb_of_instances.set_label(f"Hourly raw number of instances for {self.name}")

    def update_nb_of_instances(self):
        nb_of_instances = self.raw_nb_of_instances.ceil()

        self.nb_of_instances = nb_of_instances.set_label(f"Hourly number of instances for {self.name}")

    def update_nb_of_active_instances(self):
        nb_of_active_instances = (
                (self.storage_delta.abs() + self.storage_dumps.abs()) / self.storage_capacity
        ).to(u.dimensionless)

        self.nb_of_active_instances = nb_of_active_instances.set_label(
            f"Hourly number of active instances for {self.name}")

    def update_instances_energy(self):
        nb_of_idle_instances = (self.nb_of_instances - self.nb_of_active_instances).set_label(
            f"Hourly number of idle instances for {self.name}")
        active_storage_energy = (
                self.nb_of_active_instances * self.power * ExplainableQuantity(
            1 * u.hour, "one hour") * self.power_usage_effectiveness
        ).set_label(f"Hourly active instances energy for {self.name}")
        idle_storage_energy = (
                nb_of_idle_instances * self.idle_power * ExplainableQuantity(
            1 * u.hour, "one hour") * self.power_usage_effectiveness
        ).set_label(f"Hourly idle instances energy for {self.name}")

        storage_energy = (active_storage_energy + idle_storage_energy)

        self.instances_energy = storage_energy.set_label(f"Storage energy for {self.name}")


import requests

from efootprint.constants.sources import Source, Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.core.hardware.servers.autoscaling import Autoscaling
from efootprint.constants.units import u
from efootprint.core.hardware.servers.on_premise import OnPremise


def call_boaviztapi(url, method="GET", **kwargs):
    headers = {'accept': 'application/json'}
    if method == "GET":
        response = requests.get(url, headers=headers, **kwargs)
    elif method == "POST":
        headers["Content-Type"] = "application/json"
        response = requests.post(url, headers=headers, **kwargs)

    if response.status_code == 200:
        return response.json()
    else:
        print(f"{method} request to {url} with params {kwargs} failed with status code {response.status_code}")


def get_archetypes_and_their_configs_and_impacts():
    output_dict = {}
    for archetype in call_boaviztapi('https://api.boavizta.org/v1/server/archetypes'):
        configuration = call_boaviztapi(
            url="https://api.boavizta.org/v1/server/archetype_config", params={"archetype": archetype})
        impact = call_boaviztapi(
            url="https://api.boavizta.org/v1/server/", params={"archetype": archetype})
        if impact is None:
            print(f"No impact for archetype {archetype}")
        else:
            output_dict[archetype] = {}
            output_dict[archetype]["config"] = configuration
            output_dict[archetype]["impact"] = impact

    return output_dict


def print_archetypes_and_their_configs():
    archetypes_data = get_archetypes_and_their_configs_and_impacts()

    for archetype in archetypes_data.keys():
        config = archetypes_data[archetype]["config"]
        impact = archetypes_data[archetype]["impact"]
        if "default" in config['CPU']['core_units'].keys():
            nb_cpu_core_units = config['CPU']['core_units']['default']
        else:
            nb_cpu_core_units = impact["verbose"]['CPU-1']['core_units']['value']

        nb_ssd_units = config['SSD']["units"]['default']
        nb_hdd_units = config['HDD']["units"]['default']

        if nb_hdd_units > 0 and nb_ssd_units > 0:
            raise ValueError(
                f"Archetype {archetype} has both SSD and HDD, please check and delete this exception raising if ok")
        storage_type = "SSD"
        if nb_hdd_units > 0:
            storage_type = "HDD"
        nb_storage_units = config[storage_type]["units"]['default']

        print(
            f"{archetype}: type {config['CASE']['case_type']['default']},\n"
            f"    {config['CPU']['units']['default']} cpu units with {nb_cpu_core_units} core units,\n"
            f"    {config['RAM']['units']['default']} RAM units with {config['RAM']['capacity']['default']} GB capacity,\n"
            f"    {nb_storage_units} {storage_type} units with {config[storage_type]['capacity']['default']} GB capacity,")

        total_gwp_embedded_value = impact["impacts"]["gwp"]["embedded"]["value"]
        total_gwp_embedded_unit = impact["impacts"]["gwp"]["unit"]

        if nb_storage_units > 0:
            storage_gwp_embedded_value = impact["verbose"][f"{storage_type}-1"]["impacts"]["gwp"]["embedded"]["value"]
            storage_gwp_embedded_unit = impact["verbose"][f"{storage_type}-1"]["impacts"]["gwp"]["unit"]

            assert total_gwp_embedded_unit == storage_gwp_embedded_unit
        else:
            storage_gwp_embedded_value = 0
            storage_gwp_embedded_unit = "kg"

        average_power_value = impact["verbose"]["avg_power"]["value"]
        average_power_unit = impact["verbose"]["avg_power"]["unit"]

        print(
            f"    Impact fabrication compute: {total_gwp_embedded_value - storage_gwp_embedded_value} {total_gwp_embedded_unit},\n"
            f"    Impact fabrication storage: {storage_gwp_embedded_value} {storage_gwp_embedded_unit},\n"
            f"    Average power: {round(average_power_value, 1)} {average_power_unit}\n")


def get_cloud_server(
        provider, instance_type, average_carbon_intensity, base_efootprint_class=Autoscaling,
        lifespan=None, idle_power=None, power_usage_effectiveness=None,
        server_utilization_rate=None
        ):
    if lifespan is None:
        lifespan = SourceValue(6 * u.year, Sources.HYPOTHESIS)
    if idle_power is None:
        idle_power = SourceValue(0 * u.W, Sources.HYPOTHESIS)
    if power_usage_effectiveness is None:
        power_usage_effectiveness = SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS)
    if server_utilization_rate is None:
        server_utilization_rate = SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)

    impact_url = "https://api.boavizta.org/v1/cloud/instance"
    params = {"provider": provider, "instance_type": instance_type}
    impact_source = Source(name="Boavizta API cloud instances",
                    link=f"{impact_url}?{'&'.join([key + '=' + params[key] for key in params.keys()])}")
    impact_data = call_boaviztapi(url=impact_url, params=params)
    impacts = impact_data["impacts"]
    cpu_spec = impact_data["verbose"]["CPU-1"]
    ram_spec = impact_data["verbose"]["RAM-1"]

    average_power_value = impact_data["verbose"]["avg_power"]["value"]
    average_power_unit = impact_data["verbose"]["avg_power"]["unit"]
    use_time_ratio = impact_data["verbose"]["use_time_ratio"]["value"]

    assert average_power_unit == "W"
    assert float(use_time_ratio) == 1

    return base_efootprint_class(
        f"{provider} {instance_type} instances",
        carbon_footprint_fabrication=SourceValue(impacts["gwp"]["embedded"]["value"] * u.kg, impact_source),
        # TODO: document and challenge power calculation
        power=SourceValue(average_power_value * u.W, impact_source),
        lifespan=lifespan,
        idle_power=idle_power,
        ram=SourceValue(ram_spec["units"]["value"] * ram_spec["capacity"]["value"] * u.GB, impact_source),
        cpu_cores=SourceValue(cpu_spec["units"]["value"] * cpu_spec["core_units"]["value"] * u.core, impact_source),
        power_usage_effectiveness=power_usage_effectiveness,
        average_carbon_intensity=average_carbon_intensity,
        server_utilization_rate=server_utilization_rate)


def on_premise_server_from_config(
        name: str, nb_of_cpu_units: int, nb_of_cores_per_cpu_unit: int, nb_of_ram_units: int,
        ram_quantity_per_unit_in_gb: int, average_carbon_intensity, lifespan=None, idle_power=None,
        power_usage_effectiveness=None, server_utilization_rate=None, fixed_nb_of_instances=None):
    impact_url = "https://api.boavizta.org/v1/server/"
    params = {"verbose": "true", "archetype": "platform_compute_medium", "criteria": ["gwp"]}
    data = {"model": {"type": "rack"},
            "configuration": {"cpu": {"units": nb_of_cpu_units, "core_units": nb_of_cores_per_cpu_unit},
                              "ram": [{"units": nb_of_ram_units, "capacity": ram_quantity_per_unit_in_gb}]}}

    impact_source = Source(name="Boavizta API servers",
                           link=f"{impact_url}?{'&'.join([key + '=' + str(params[key]) for key in params.keys()])}")
    impact_data = call_boaviztapi(url=impact_url, params=params, json=data, method="POST")

    impacts = impact_data["impacts"]
    cpu_spec = impact_data["verbose"]["CPU-1"]
    ram_spec = impact_data["verbose"]["RAM-1"]

    if lifespan is None:
        lifespan = SourceValue(6 * u.year, Sources.HYPOTHESIS)
    if idle_power is None:
        idle_power = SourceValue(0 * u.W, Sources.HYPOTHESIS)
    if power_usage_effectiveness is None:
        power_usage_effectiveness = SourceValue(1.4 * u.dimensionless, Sources.HYPOTHESIS)
    if server_utilization_rate is None:
        server_utilization_rate = SourceValue(0.7 * u.dimensionless, Sources.HYPOTHESIS)

    average_power_value = impact_data["verbose"]["avg_power"]["value"]
    average_power_unit = impact_data["verbose"]["avg_power"]["unit"]
    use_time_ratio = impact_data["verbose"]["use_time_ratio"]["value"]

    assert average_power_unit == "W"
    assert float(use_time_ratio) == 1

    return OnPremise(
        name,
        carbon_footprint_fabrication=SourceValue(impacts["gwp"]["embedded"]["value"] * u.kg, impact_source),
        # TODO: document and challenge power calculation
        power=SourceValue(average_power_value * u.W, impact_source),
        lifespan=lifespan,
        idle_power=idle_power,
        ram=SourceValue(ram_spec["units"]["value"] * ram_spec["capacity"]["value"] * u.GB, impact_source),
        cpu_cores=SourceValue(cpu_spec["units"]["value"] * cpu_spec["core_units"]["value"] * u.core, impact_source),
        power_usage_effectiveness=power_usage_effectiveness,
        average_carbon_intensity=average_carbon_intensity,
        server_utilization_rate=server_utilization_rate,
        fixed_nb_of_instances=fixed_nb_of_instances)


if __name__ == "__main__":
    print_archetypes_and_their_configs()



from pyvis.network import Network

from efootprint.utils.graph_tools import WIDTH, HEIGHT, set_string_max_width

COLOR_MAP = {
    "Autoscaling": "red",
    "OnPremise": "red",
    "Serverless": "red",
    "Hardware": "red",
    "Storage": "red",
    "Service": "gold",
    "UsagePattern": "red",
    "UserJourney": "dodgerblue",
    "UserJourneyStep": "deepskyblue",
    "TimeIntervals": "gold",
    "Job": "palegoldenrod"
}

USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE = [
    "System", "Network", "Hardware", "Country"]
SERVICES_VIEW_CLASSES_TO_IGNORE = [
    "System", "UsagePattern", "TimeIntervals", "Network", "Server", "Storage", "Hardware"]
SERVICES_AND_INFRA_VIEW_CLASSES_TO_IGNORE = [
    "UsagePattern", "TimeIntervals", "Network", "Hardware", "System", "UserJourney", "UserJourneyStep"]


def build_object_relationships_graph(
        node, input_graph=None, visited=None, classes_to_ignore=None, width=WIDTH, height=HEIGHT, notebook=False):
    cdn_resources = "local"
    if notebook:
        cdn_resources = "in_line"
    if classes_to_ignore is None:
        classes_to_ignore = ["System"]
    if input_graph is None:
        input_graph = Network(notebook=notebook, width=width, height=height, cdn_resources=cdn_resources)
    if visited is None:
        visited = set()

    if node in visited:
        return input_graph

    node_type = type(node).__name__
    if node_type not in classes_to_ignore:
        input_graph.add_node(
            id(node), label=set_string_max_width(f"{node.name}", 20),
            title=set_string_max_width(str(node), 80),
            color=COLOR_MAP.get(node_type, "gray"))

    for mod_obj in node.mod_obj_attributes:
        mod_obj_type = type(mod_obj).__name__
        if mod_obj_type not in classes_to_ignore:
            input_graph.add_node(
                id(mod_obj), label=set_string_max_width(f"{mod_obj.name}", 20),
                title=set_string_max_width(str(mod_obj), 80),
                color=COLOR_MAP.get(mod_obj_type, "gray"))
            if node_type not in classes_to_ignore:
                input_graph.add_edge(id(node), id(mod_obj))

        if mod_obj not in visited:
            visited.add(node)
            build_object_relationships_graph(mod_obj, input_graph, visited, classes_to_ignore, width, height)

    return input_graph



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.servers.autoscaling import Autoscaling
from efootprint.core.hardware.servers.on_premise import OnPremise
from efootprint.core.hardware.servers.serverless import Serverless


def default_serverless(name="Default serverless", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(300 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(6 * u.year, Sources.HYPOTHESIS),
        "idle_power": SourceValue(50 * u.W, Sources.HYPOTHESIS),
        "ram": SourceValue(128 * u.GB, Sources.HYPOTHESIS),
        "cpu_cores": SourceValue(24 * u.core, Sources.HYPOTHESIS),
        "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
        "average_carbon_intensity": SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
        "server_utilization_rate": SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return Serverless(name, **output_args)


def default_autoscaling(name="Default autoscaling", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(300 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(6 * u.year, Sources.HYPOTHESIS),
        "idle_power": SourceValue(50 * u.W, Sources.HYPOTHESIS),
        "ram": SourceValue(128 * u.GB, Sources.HYPOTHESIS),
        "cpu_cores": SourceValue(24 * u.core, Sources.HYPOTHESIS),
        "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
        "average_carbon_intensity": SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
        "server_utilization_rate": SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return Autoscaling(name, **output_args)


def default_onpremise(name="Default on premise", **kwargs):
    output_args = {
        "carbon_footprint_fabrication": SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
        "power": SourceValue(300 * u.W, Sources.HYPOTHESIS),
        "lifespan": SourceValue(6 * u.year, Sources.HYPOTHESIS),
        "idle_power": SourceValue(50 * u.W, Sources.HYPOTHESIS),
        "ram": SourceValue(128 * u.GB, Sources.HYPOTHESIS),
        "cpu_cores": SourceValue(24 * u.core, Sources.HYPOTHESIS),
        "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
        "average_carbon_intensity": SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
        "server_utilization_rate": SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
    }

    output_args.update(kwargs)

    return OnPremise(name, **output_args)



import time


def round_dict(my_dict, round_level):
    for key in my_dict:
        my_dict[key] = round(my_dict[key], round_level)

    return my_dict


def flatten_list(nested_list):
    flattened_list = [item for sublist in nested_list
                      for item in (flatten_list(sublist) if isinstance(sublist, list) else [sublist])]
    for elt in flattened_list:
        if type(elt) == set:
            raise ValueError("flatten_list function is not supposed to handle set elements, please refactor it.")

    return flattened_list


def time_it(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        diff = end_time - start_time
        if diff > 0.001:
            print(f"Function {func.__name__} took {diff:.5f} seconds to execute.")
        return result
    return wrapper


def convert_to_list(input_value):
    if type(input_value) == list:
        value_elts = flatten_list(input_value)
    elif type(input_value) == set:
        value_elts = flatten_list(list(input_value))
    else:
        value_elts = [input_value]

    return value_elts


def format_co2_amount(co2_amount_in_kg: int, rounding_value=1):
    if co2_amount_in_kg < 501:
        unit = "kg"
        dividing_number = 1
    else:
        unit = "tonne"
        dividing_number = 1000
    rounded_total__new = round(co2_amount_in_kg / dividing_number, rounding_value)
    if rounding_value == 0:
        rounded_total__new = int(rounded_total__new)

    return rounded_total__new, unit


def display_co2_amount(num_value_and_unit_tuple):
    num_value, unit = num_value_and_unit_tuple

    return f"{num_value} {unit}s"
